import copy

import numpy as np

import autofit as af
from autolens import exc
from autolens.lens import lens_data as ld, lens_fit
from autolens.model.galaxy import galaxy as g
from autolens.model.inversion import pixelizations as px
from autolens.model.inversion import regularization as rg
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline.phase.phase import setup_phase_mask


class HyperPixelizationPhase(phase_imaging.PhaseImaging, af.HyperPhase):
    """
    Phase that makes everything in the variable from the previous phase equal to the
    corresponding value from the best fit except for variables associated with
    pixelization
    """

    def run(self, data, results=None, mask=None, positions=None):
        """
        Run the phase, overriding the optimizer's variable instance with one created to
        only fit pixelization hyperparameters.
        """
        variable = copy.deepcopy(results.last.variable)
        HyperPixelizationPhase.transfer_classes(results.last.constant, variable)
        self.optimizer.variable = variable
        new_result = super().run(data, results=results, mask=mask, positions=positions)
        result = results.last
        result.hyper = new_result
        return result

    @staticmethod
    def transfer_classes(instance, mapper):
        """
        Recursively overwrite priors in the mapper with constant values from the
        instance except where the containing class is associated with pixelization.
        Parameters
        ----------
        instance
            The best fit from the previous phase
        mapper
            The prior variable from the previous phase
        """
        for key, instance_value in instance.__dict__.items():
            try:
                mapper_value = getattr(mapper, key)
                if isinstance(mapper_value, af.Prior):
                    setattr(mapper, key, instance_value)
                if not (isinstance(instance_value, px.Pixelization) or isinstance(
                        instance_value, rg.Regularization)):
                    try:
                        HyperPixelizationPhase.transfer_classes(instance_value,
                                                                mapper_value)
                    except AttributeError:
                        setattr(mapper, key, instance_value)
            except AttributeError:
                pass

    @property
    def uses_inversion(self):
        return True

    @property
    def uses_hyper_images(self):
        return True

    class Analysis(phase_imaging.LensSourcePlanePhase.Analysis):

        def figure_of_merit_for_fit(self, tracer):
            pass

        def __init__(self, lens_data, cosmology, positions_threshold, results=None,
                     uses_hyper_images=False):
            super(HyperPixelizationPhase.Analysis, self).__init__(
                lens_data=lens_data, cosmology=cosmology,
                positions_threshold=positions_threshold,
                results=results, uses_hyper_images=uses_hyper_images)


class HyperGalaxyPhase(phase_imaging.PhaseImaging, af.HyperPhase):

    class Analysis(af.Analysis):

        def __init__(self, lens_data, model_image_2d, galaxy_image_2d):
            """
            An analysis to fit the noise for a single galaxy image.
            Parameters
            ----------
            lens_data: LensData
                Lens data, including an image and noise
            model_image_2d: ndarray
                An image produce of the overall system by a model
            galaxy_image_2d: ndarray
                The contribution of one galaxy to the model image
            """
            self.lens_data = lens_data
            self.hyper_model_image_1d = lens_data.array_1d_from_array_2d(
                array_2d=model_image_2d)
            self.hyper_galaxy_image_1d = lens_data.array_1d_from_array_2d(
                array_2d=galaxy_image_2d)

            self.check_for_previously_masked_values(array=self.hyper_model_image_1d)
            self.check_for_previously_masked_values(array=self.hyper_galaxy_image_1d)

        @staticmethod
        def check_for_previously_masked_values(array):
            if not np.all(array) != 0.0:
                raise exc.PhaseException(
                    'When mapping a 2D array to a 1D array using lens data, a value '
                    'encountered was 0.0 and therefore masked in a previous phase.')

        def visualize(self, instance, image_path, during_analysis):
            pass

        def fit(self, instance):
            """
            Fit the model image to the real image by scaling the hyper noise.
            Parameters
            ----------
            instance: ModelInstance
                A model instance with a hyper galaxy property
            Returns
            -------
            fit: float
            """
            fit = self.fit_for_hyper_galaxy(hyper_galaxy=instance.hyper_galaxy)
            return fit.figure_of_merit

        def fit_for_hyper_galaxy(self, hyper_galaxy):
            hyper_noise_1d = (
                hyper_galaxy.hyper_noise_map_from_hyper_images_and_noise_map(
                    hyper_model_image=self.hyper_model_image_1d,
                    hyper_galaxy_image=self.hyper_galaxy_image_1d,
                    noise_map=self.lens_data.noise_map_1d
                )
            )

            hyper_noise_map_1d = self.lens_data.noise_map_1d + hyper_noise_1d
            return lens_fit.LensDataFit(
                image_1d=self.lens_data.image_1d,
                noise_map_1d=hyper_noise_map_1d,
                mask_1d=self.lens_data.mask_1d,
                model_image_1d=self.hyper_model_image_1d,
                map_to_scaled_array=self.lens_data.map_to_scaled_array
            )

        @classmethod
        def describe(cls, instance):
            return "Running hyper galaxy fit for HyperGalaxy:\n{}".format(
                instance.hyper_galaxy)

    def run(self, data, results=None, mask=None, positions=None):
        """
        Run a fit for each galaxy from the previous phase.
        Parameters
        ----------
        data: LensData
        results: ResultsCollection
            Results from all previous phases
        mask: Mask
            The mask
        positions
        Returns
        -------
        results: HyperGalaxyResults
            A collection of results, with one item per a galaxy
        """

        mask = setup_phase_mask(data=data, mask=mask, mask_function=self.mask_function,
                                inner_mask_radii=self.inner_mask_radii)

        lens_data = ld.LensData(ccd_data=data, mask=mask,
                                sub_grid_size=self.sub_grid_size,
                                image_psf_shape=self.image_psf_shape,
                                positions=positions,
                                interp_pixel_scale=self.interp_pixel_scale,
                                uses_inversion=self.uses_inversion)

        model_image_2d = results.last.most_likely_fit.model_image_2d

        hyper_result = copy.deepcopy(results.last)
        hyper_result.analysis.uses_hyper_images = True
        hyper_result.analysis.hyper_model_image_1d = lens_data.array_1d_from_array_2d(
            array_2d=model_image_2d)
        hyper_result.analysis.hyper_galaxy_image_1d_path_dict = {}

        for galaxy_path, galaxy in results.last.path_galaxy_tuples:

            optimizer = self.optimizer.copy_with_name_extension(
                extension=galaxy_path[-1])
            optimizer.variable.hyper_galaxy = g.HyperGalaxy
            galaxy_image_2d = results.last.image_2d_dict[galaxy_path]

            # If array is all zeros, galaxy did not have image in previous phase and
            # should be ignored
            if not np.all(galaxy_image_2d == 0):
                hyper_result.analysis.hyper_galaxy_image_1d_path_dict[
                    galaxy_path] = lens_data.array_1d_from_array_2d(
                    array_2d=galaxy_image_2d)
                analysis = self.__class__.Analysis(lens_data=lens_data,
                                                   model_image_2d=model_image_2d,
                                                   galaxy_image_2d=galaxy_image_2d)
                result = optimizer.fit(analysis)

                hyper_result.constant.object_for_path(
                    galaxy_path
                ).hyper_galaxy = result.constant.hyper_galaxy

        results.hyper = hyper_result

        return results



import numpy as np
from astropy import cosmology as cosmo

import autofit as af
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy as g, galaxy_fit, galaxy_data as gd
from autolens.model.galaxy.plotters import galaxy_fit_plotters


def default_mask_function(image):
    return msk.Mask.circular(shape=image.shape, pixel_scale=image.pixel_scale,
                             radius_arcsec=3.0)


def setup_phase_mask(data, mask, mask_function, inner_mask_radii):
    if mask_function is not None:
        mask = mask_function(image=data.image)
    elif mask is None and mask_function is None:
        mask = default_mask_function(image=data.image)

    if inner_mask_radii is not None:
        inner_mask = msk.Mask.circular(shape=mask.shape, pixel_scale=mask.pixel_scale,
                                       radius_arcsec=inner_mask_radii, invert=True)
        mask = mask + inner_mask

    return mask


class AbstractPhase(af.AbstractPhase):

    def __init__(self, phase_name, phase_tag=None, phase_folders=None, tag_phases=True,
                 optimizer_class=af.MultiNest,
                 cosmology=cosmo.Planck15, auto_link_priors=False):
        """
        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit
        models and hyper passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        phase_name: str
            The name of this phase
        """

        super().__init__(phase_name=phase_name, phase_tag=phase_tag,
                         phase_folders=phase_folders, tag_phases=tag_phases,
                         optimizer_class=optimizer_class,
                         auto_link_priors=auto_link_priors)

        self.cosmology = cosmology

    @property
    def variable(self):
        """
        Convenience method

        Returns
        -------
        ModelMapper
            A model mapper comprising all the variable (prior) objects in this lens
        """
        return self.optimizer.variable

    @property
    def phase_property_collections(self):
        """
        Returns
        -------
        phase_property_collections: [PhaseProperty]
            A list of phase property collections associated with this phase. This is
            used in automated prior passing and should be overridden for any phase that
            contains its own PhasePropertys.
        """
        return []

    @property
    def path(self):
        return self.optimizer.path

    @property
    def doc(self):
        if self.__doc__ is not None:
            return self.__doc__.replace("  ", "").replace("\n", " ")

    def pass_priors(self, results):
        """
        Perform any prior or constant passing. This could involve setting model
        attributes equal to priors or constants from a previous phase.

        Parameters
        ----------
        results: autofit.tools.pipeline.ResultsCollection
            The result of the previous phase
        """
        pass

    # noinspection PyAbstractClass
    class Analysis(af.Analysis):

        def __init__(self, cosmology, results=None):
            """
            An lens object

            Parameters
            ----------
            results: autofit.tools.pipeline.ResultsCollection
                The results of all previous phases
            """

            self.results = results
            self.cosmology = cosmology

            self.plot_count = 0

        @property
        def last_results(self):
            """
            Returns
            -------
            result: AbstractPhase.Result | None
                The result from the last phase
            """
            if self.results is not None:
                return self.results.last

        def tracer_for_instance(self, instance):
            raise NotImplementedError()

        def padded_tracer_for_instance(self, instance):
            raise NotImplementedError()

        def fit_for_tracers(self, tracer, padded_tracer):
            raise NotImplementedError()

        def figure_of_merit_for_fit(self, tracer):
            raise NotImplementedError()

    def make_result(self, result, analysis):
        return self.__class__.Result(constant=result.constant,
                                     figure_of_merit=result.figure_of_merit,
                                     previous_variable=result.previous_variable,
                                     gaussian_tuples=result.gaussian_tuples,
                                     analysis=analysis, optimizer=self.optimizer)

    class Result(af.Result):

        def __init__(self, constant, figure_of_merit, previous_variable,
                     gaussian_tuples, analysis, optimizer):
            """
            The result of a phase
            """
            super(Phase.Result, self).__init__(constant=constant,
                                               figure_of_merit=figure_of_merit,
                                               previous_variable=previous_variable,
                                               gaussian_tuples=gaussian_tuples)

            self.analysis = analysis
            self.optimizer = optimizer

        @property
        def most_likely_tracer(self):
            return self.analysis.tracer_for_instance(instance=self.constant)

        @property
        def most_likely_padded_tracer(self):
            return self.analysis.padded_tracer_for_instance(instance=self.constant)

        @property
        def most_likely_fit(self):
            return self.analysis.fit_for_tracers(
                tracer=self.most_likely_tracer,
                padded_tracer=self.most_likely_padded_tracer)

        @property
        def unmasked_model_image(self):
            return self.most_likely_fit.unmasked_blurred_image_plane_image

        @property
        def unmasked_model_image_of_planes(self):
            return self.most_likely_fit.unmasked_blurred_image_plane_image_of_planes

        @property
        def unmasked_model_image_of_planes_and_galaxies(self):
            fit = self.most_likely_fit
            return fit.unmasked_blurred_image_plane_image_of_planes_and_galaxies

        def image_2d_for_galaxy(self, galaxy: g.Galaxy) -> np.ndarray:
            """
            Parameters
            ----------
            galaxy
                A galaxy used in this phase

            Returns
            -------
            ndarray or None
                A numpy array giving the model image of that galaxy
            """
            return self.most_likely_fit.galaxy_image_2d_dict[galaxy]

        @property
        def path_galaxy_tuples(self) -> [(str, g.Galaxy)]:
            """
            Tuples associating the names of galaxies with instances from the best fit
            """
            return self.constant.path_instance_tuples_for_class(cls=g.Galaxy)

        @property
        def image_2d_dict(self) -> {str: g.Galaxy}:
            """
            A dictionary associating galaxy names with model images of those galaxies
            """
            return {galaxy_path: self.image_2d_for_galaxy(galaxy) for
                    galaxy_path, galaxy in self.path_galaxy_tuples}


class Phase(AbstractPhase):

    def run(self, image, results=None, mask=None):
        raise NotImplementedError()

    # noinspection PyAbstractClass
    class Analysis(AbstractPhase.Analysis):

        def __init__(self, cosmology, results=None):
            super(Phase.Analysis, self).__init__(cosmology=cosmology, results=results)

            self.should_plot_mask = \
                af.conf.instance.general.get('output', 'plot_mask_on_images', bool)
            self.extract_array_from_mask = \
                af.conf.instance.general.get('output', 'extract_images_from_mask', bool)
            self.zoom_around_mask = \
                af.conf.instance.general.get('output', 'zoom_around_mask_of_images', bool)
            self.should_plot_positions = \
                af.conf.instance.general.get('output', 'plot_positions_on_images', bool)
            self.plot_units = \
                af.conf.instance.general.get('output', 'plot_units', str).strip()

            self.plot_ray_tracing_all_at_end_png = \
                af.conf.instance.general.get('output', 'plot_ray_tracing_all_at_end_png',
                                          bool)
            self.plot_ray_tracing_all_at_end_fits = \
                af.conf.instance.general.get('output', 'plot_ray_tracing_all_at_end_fits',
                                          bool)

            self.plot_ray_tracing_as_subplot = \
                af.conf.instance.general.get('output', 'plot_ray_tracing_as_subplot', bool)
            self.plot_ray_tracing_image_plane_image = \
                af.conf.instance.general.get('output',
                                          'plot_ray_tracing_image_plane_image', bool)
            self.plot_ray_tracing_source_plane = \
                af.conf.instance.general.get('output',
                                          'plot_ray_tracing_source_plane_image', bool)
            self.plot_ray_tracing_convergence = \
                af.conf.instance.general.get('output', 'plot_ray_tracing_convergence',
                                          bool)
            self.plot_ray_tracing_potential = \
                af.conf.instance.general.get('output', 'plot_ray_tracing_potential', bool)
            self.plot_ray_tracing_deflections = \
                af.conf.instance.general.get('output', 'plot_ray_tracing_deflections',
                                          bool)


class GalaxyFitPhase(AbstractPhase):
    galaxies = af.PhaseProperty("galaxies")

    def __init__(self, phase_name, tag_phases=True, phase_folders=None, galaxies=None,
                 use_intensities=False,
                 use_convergence=False,
                 use_potential=False,
                 use_deflections=False, optimizer_class=af.MultiNest,
                 sub_grid_size=2, interp_pixel_scale=None,
                 mask_function=None, cosmology=cosmo.Planck15):
        """
        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit
        models and hyper passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """

        super(GalaxyFitPhase, self).__init__(phase_name=phase_name,
                                             tag_phases=tag_phases,
                                             phase_folders=phase_folders,
                                             optimizer_class=optimizer_class,
                                             cosmology=cosmology)
        self.use_intensities = use_intensities
        self.use_convergence = use_convergence
        self.use_potential = use_potential
        self.use_deflections = use_deflections
        self.galaxies = galaxies
        self.sub_grid_size = sub_grid_size
        self.interp_pixel_scale = interp_pixel_scale
        self.mask_function = mask_function

    def run(self, galaxy_data, results=None, mask=None):
        """
        Run this phase.

        Parameters
        ----------
        galaxy_data
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has
            been executed

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper.
        """
        analysis = self.make_analysis(galaxy_data=galaxy_data, results=results,
                                      mask=mask)

        self.pass_priors(results)
        self.assert_and_save_pickle()

        result = self.run_analysis(analysis)

        return self.make_result(result, analysis)

    def make_analysis(self, galaxy_data, results=None, mask=None):
        """
        Create an lens object. Also calls the prior passing and lens_data modifying
        functions to allow child classes to change the behaviour of the phase.

        Parameters
        ----------
        galaxy_data
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens: Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a
             set of values
        """

        mask = setup_phase_mask(data=galaxy_data[0], mask=mask,
                                mask_function=self.mask_function,
                                inner_mask_radii=None)

        if self.use_intensities or self.use_convergence or self.use_potential:

            galaxy_data = gd.GalaxyFitData(galaxy_data=galaxy_data[0], mask=mask,
                                           sub_grid_size=self.sub_grid_size,
                                           interp_pixel_scale=self.interp_pixel_scale,
                                           use_intensities=self.use_intensities,
                                           use_convergence=self.use_convergence,
                                           use_potential=self.use_potential,
                                           use_deflections_y=self.use_deflections,
                                           use_deflections_x=self.use_deflections)

            return self.__class__.AnalysisSingle(galaxy_data=galaxy_data,
                                                 cosmology=self.cosmology,
                                                 results=results)

        elif self.use_deflections:

            galaxy_data_y = gd.GalaxyFitData(galaxy_data=galaxy_data[0], mask=mask,
                                             sub_grid_size=self.sub_grid_size,
                                             interp_pixel_scale=self.interp_pixel_scale,
                                             use_intensities=self.use_intensities,
                                             use_convergence=self.use_convergence,
                                             use_potential=self.use_potential,
                                             use_deflections_y=self.use_deflections,
                                             use_deflections_x=False)

            galaxy_data_x = gd.GalaxyFitData(galaxy_data=galaxy_data[1], mask=mask,
                                             sub_grid_size=self.sub_grid_size,
                                             interp_pixel_scale=self.interp_pixel_scale,
                                             use_intensities=self.use_intensities,
                                             use_convergence=self.use_convergence,
                                             use_potential=self.use_potential,
                                             use_deflections_y=False,
                                             use_deflections_x=self.use_deflections)

            return self.__class__.AnalysisDeflections(galaxy_data_y=galaxy_data_y,
                                                      galaxy_data_x=galaxy_data_x,
                                                      cosmology=self.cosmology,
                                                      results=results)

    # noinspection PyAbstractClass
    class Analysis(Phase.Analysis):

        def __init__(self, cosmology, results):
            super(GalaxyFitPhase.Analysis, self).__init__(cosmology=cosmology,
                                                          results=results)

            self.plot_galaxy_fit_all_at_end_png = \
                af.conf.instance.general.get('output', 'plot_galaxy_fit_all_at_end_png',
                                          bool)
            self.plot_galaxy_fit_all_at_end_fits = \
                af.conf.instance.general.get('output', 'plot_galaxy_fit_all_at_end_fits',
                                          bool)
            self.plot_galaxy_fit_as_subplot = \
                af.conf.instance.general.get('output', 'plot_galaxy_fit_as_subplot', bool)
            self.plot_galaxy_fit_image = \
                af.conf.instance.general.get('output', 'plot_galaxy_fit_image', bool)
            self.plot_galaxy_fit_noise_map = \
                af.conf.instance.general.get('output', 'plot_galaxy_fit_noise_map', bool)
            self.plot_galaxy_fit_model_image = \
                af.conf.instance.general.get('output', 'plot_galaxy_fit_model_image', bool)
            self.plot_galaxy_fit_residual_map = \
                af.conf.instance.general.get('output', 'plot_galaxy_fit_residual_map',
                                          bool)
            self.plot_galaxy_fit_chi_squared_map = \
                af.conf.instance.general.get('output', 'plot_galaxy_fit_chi_squared_map',
                                          bool)

        @classmethod
        def describe(cls, instance):
            return "\nRunning galaxy fit for... \n\nGalaxies::\n{}\n\n".format(
                instance.galaxies)

    # noinspection PyAbstractClass
    class AnalysisSingle(Analysis):

        def __init__(self, galaxy_data, cosmology, results=None):
            super(GalaxyFitPhase.AnalysisSingle, self).__init__(cosmology=cosmology,
                                                                results=results)

            self.galaxy_data = galaxy_data

        def fit(self, instance):
            fit = self.fit_for_instance(instance=instance)
            return fit.figure_of_merit

        def visualize(self, instance, image_path, during_analysis):

            self.plot_count += 1
            fit = self.fit_for_instance(instance=instance)

            if self.plot_galaxy_fit_as_subplot:
                galaxy_fit_plotters.plot_fit_subplot(
                    fit=fit, should_plot_mask=self.should_plot_mask,
                    zoom_around_mask=self.zoom_around_mask,
                    units=self.plot_units,
                    output_path=image_path, output_format='png')

            if during_analysis:

                galaxy_fit_plotters.plot_fit_individuals(
                    fit=fit, should_plot_mask=self.should_plot_mask,
                    zoom_around_mask=self.zoom_around_mask,
                    should_plot_image=self.plot_galaxy_fit_image,
                    should_plot_noise_map=self.plot_galaxy_fit_noise_map,
                    should_plot_model_image=self.plot_galaxy_fit_model_image,
                    should_plot_residual_map=self.plot_galaxy_fit_residual_map,
                    should_plot_chi_squared_map=self.plot_galaxy_fit_chi_squared_map,
                    units=self.plot_units,
                    output_path=image_path, output_format='png')

            elif not during_analysis:

                if self.plot_ray_tracing_all_at_end_png:
                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit, should_plot_mask=self.should_plot_mask,
                        zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=image_path, output_format='png')

                if self.plot_ray_tracing_all_at_end_fits:
                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit, should_plot_mask=self.should_plot_mask,
                        zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path="{}/fits/".format(image_path), output_format='fits')

            return fit

        def fit_for_instance(self, instance):
            """
            Determine the fit of a lens galaxy and source galaxy to the lens_data in
            this lens.

            Parameters
            ----------
            instance
                A model instance with attributes

            Returns
            -------
            fit: Fit
                A fractional value indicating how well this model fit and the model
                lens_data itself
            """
            return galaxy_fit.GalaxyFit(galaxy_data=self.galaxy_data,
                                        model_galaxies=instance.galaxies)

    # noinspection PyAbstractClass
    class AnalysisDeflections(Analysis):

        def __init__(self, galaxy_data_y, galaxy_data_x, cosmology, results=None):
            super(GalaxyFitPhase.AnalysisDeflections, self).__init__(
                cosmology=cosmology,
                results=results)

            self.galaxy_data_y = galaxy_data_y
            self.galaxy_data_x = galaxy_data_x

        def fit(self, instance):
            fit_y, fit_x = self.fit_for_instance(instance=instance)
            return fit_y.figure_of_merit + fit_x.figure_of_merit

        def visualize(self, instance, image_path, during_analysis):

            output_image_y_path = "{}/fit_y_".format(image_path)
            output_fits_y_path = "{}/fits/fit_y".format(image_path)
            output_image_x_path = "{}/fit_x_".format(image_path)
            output_fits_x_path = "{}/fits/fit_x".format(image_path)

            self.plot_count += 1
            fit_y, fit_x = self.fit_for_instance(instance=instance)

            if self.plot_galaxy_fit_as_subplot:
                galaxy_fit_plotters.plot_fit_subplot(
                    fit=fit_y, should_plot_mask=self.should_plot_mask,
                    zoom_around_mask=self.zoom_around_mask,
                    units=self.plot_units,
                    output_path=output_image_y_path, output_format='png')

                galaxy_fit_plotters.plot_fit_subplot(
                    fit=fit_x, should_plot_mask=self.should_plot_mask,
                    zoom_around_mask=self.zoom_around_mask,
                    units=self.plot_units,
                    output_path=output_image_x_path, output_format='png')

            if during_analysis:

                galaxy_fit_plotters.plot_fit_individuals(
                    fit=fit_y, should_plot_mask=self.should_plot_mask,
                    zoom_around_mask=self.zoom_around_mask,
                    should_plot_image=self.plot_galaxy_fit_image,
                    should_plot_noise_map=self.plot_galaxy_fit_noise_map,
                    should_plot_model_image=self.plot_galaxy_fit_model_image,
                    should_plot_residual_map=self.plot_galaxy_fit_residual_map,
                    should_plot_chi_squared_map=self.plot_galaxy_fit_chi_squared_map,
                    units=self.plot_units,
                    output_path=output_image_y_path, output_format='png')

                galaxy_fit_plotters.plot_fit_individuals(
                    fit=fit_x, should_plot_mask=self.should_plot_mask,
                    zoom_around_mask=self.zoom_around_mask,
                    should_plot_image=self.plot_galaxy_fit_image,
                    should_plot_noise_map=self.plot_galaxy_fit_noise_map,
                    should_plot_model_image=self.plot_galaxy_fit_model_image,
                    should_plot_residual_map=self.plot_galaxy_fit_residual_map,
                    should_plot_chi_squared_map=self.plot_galaxy_fit_chi_squared_map,
                    units=self.plot_units,
                    output_path=output_image_x_path, output_format='png')

            elif not during_analysis:

                if self.plot_ray_tracing_all_at_end_png:
                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit_y, should_plot_mask=self.should_plot_mask,
                        zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=output_image_y_path, output_format='png')

                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit_x, should_plot_mask=self.should_plot_mask,
                        zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=output_image_x_path, output_format='png')

                if self.plot_ray_tracing_all_at_end_fits:
                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit_y, should_plot_mask=self.should_plot_mask,
                        zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=output_fits_y_path, output_format='fits')

                    galaxy_fit_plotters.plot_fit_individuals(
                        fit=fit_x, should_plot_mask=self.should_plot_mask,
                        zoom_around_mask=self.zoom_around_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_model_image=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        units=self.plot_units,
                        output_path=output_fits_x_path, output_format='fits')

            return fit_y, fit_x

        def fit_for_instance(self, instance):

            fit_y = galaxy_fit.GalaxyFit(galaxy_data=self.galaxy_data_y,
                                         model_galaxies=instance.galaxies)
            fit_x = galaxy_fit.GalaxyFit(galaxy_data=self.galaxy_data_x,
                                         model_galaxies=instance.galaxies)

            return fit_y, fit_x



from autolens.lens.plotters import ray_tracing_plotters
from test.fixtures import *


@pytest.fixture(name='ray_tracing_plotter_path')
def make_ray_tracing_plotter_setup():
    return "{}/../../test_files/plotting/ray_tracing/".format(os.path.dirname(os.path.realpath(__file__)))


def test__image_plane_image_is_output(tracer_x2_plane_5x5, mask_5x5, ray_tracing_plotter_path, plot_patch):

    ray_tracing_plotters.plot_image_plane_image(tracer=tracer_x2_plane_5x5, mask=mask_5x5, extract_array_from_mask=True,
                                                zoom_around_mask=True, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                                output_path=ray_tracing_plotter_path, output_format='png')

    assert ray_tracing_plotter_path + 'tracer_image_plane_image.png' in plot_patch.paths


def test__convergence_is_output(tracer_x2_plane_5x5, mask_5x5, ray_tracing_plotter_path, plot_patch):

    ray_tracing_plotters.plot_convergence(tracer=tracer_x2_plane_5x5, mask=mask_5x5, extract_array_from_mask=True,
                                          zoom_around_mask=True, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                          output_path=ray_tracing_plotter_path, output_format='png')

    assert ray_tracing_plotter_path + 'tracer_convergence.png' in plot_patch.paths


def test__potential_is_output(tracer_x2_plane_5x5, mask_5x5, ray_tracing_plotter_path, plot_patch):

    ray_tracing_plotters.plot_potential(tracer=tracer_x2_plane_5x5, mask=mask_5x5, extract_array_from_mask=True,
                                        zoom_around_mask=True, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                        output_path=ray_tracing_plotter_path, output_format='png')

    assert ray_tracing_plotter_path + 'tracer_potential.png' in plot_patch.paths


def test__deflections_y_is_output(tracer_x2_plane_5x5, mask_5x5, ray_tracing_plotter_path, plot_patch):

    ray_tracing_plotters.plot_deflections_y(tracer=tracer_x2_plane_5x5, mask=mask_5x5, extract_array_from_mask=True,
                                            zoom_around_mask=True, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                            output_path=ray_tracing_plotter_path, output_format='png')

    assert ray_tracing_plotter_path + 'tracer_deflections_y.png' in plot_patch.paths


def test__deflections_x_is_output(tracer_x2_plane_5x5, mask_5x5, ray_tracing_plotter_path, plot_patch):

    ray_tracing_plotters.plot_deflections_x(tracer=tracer_x2_plane_5x5, mask=mask_5x5, extract_array_from_mask=True,
                                            zoom_around_mask=True, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                            output_path=ray_tracing_plotter_path, output_format='png')

    assert ray_tracing_plotter_path + 'tracer_deflections_x.png' in plot_patch.paths


def test__tracer_sub_plot_output(tracer_x2_plane_5x5, mask_5x5, ray_tracing_plotter_path, plot_patch):

    ray_tracing_plotters.plot_ray_tracing_subplot(
        tracer=tracer_x2_plane_5x5, mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
        output_path=ray_tracing_plotter_path, output_format='png')

    assert ray_tracing_plotter_path + 'tracer.png' in plot_patch.paths


def test__tracer_individuals__dependent_on_input(tracer_x2_plane_5x5, mask_5x5, ray_tracing_plotter_path, plot_patch):

    ray_tracing_plotters.plot_ray_tracing_individual(
        tracer=tracer_x2_plane_5x5, mask=mask_5x5, extract_array_from_mask=True, zoom_around_mask=True,
        should_plot_image_plane_image=True, should_plot_source_plane=True, should_plot_potential=True,
        output_path=ray_tracing_plotter_path, output_format='png')

    assert ray_tracing_plotter_path + 'tracer_image_plane_image.png' in plot_patch.paths

    assert ray_tracing_plotter_path + 'tracer_source_plane.png' in plot_patch.paths

    assert ray_tracing_plotter_path + 'tracer_convergence.png' not in plot_patch.paths

    assert ray_tracing_plotter_path + 'tracer_potential.png' in plot_patch.paths

    assert ray_tracing_plotter_path + 'tracer_deflections_y.png' not in plot_patch.paths

    assert ray_tracing_plotter_path + 'tracer_deflections_x.png' not in plot_patch.paths


def test__plot_ray_tracing_for_phase__dependent_on_input(tracer_x2_plane_5x5, mask_5x5, ray_tracing_plotter_path, plot_patch):

    ray_tracing_plotters.plot_ray_tracing_for_phase(
        tracer=tracer_x2_plane_5x5, during_analysis=False, mask=mask_5x5, positions=None,
        extract_array_from_mask=True, zoom_around_mask=True, units='arcsec',
        should_plot_as_subplot=True,
        should_plot_all_at_end_png=False,
        should_plot_all_at_end_fits=False,
        should_plot_image_plane_image=True,
        should_plot_source_plane=True,
        should_plot_convergence=False,
        should_plot_potential=True,
        should_plot_deflections=False,
        visualize_path=ray_tracing_plotter_path)

    assert ray_tracing_plotter_path + 'tracer.png' in plot_patch.paths
    assert ray_tracing_plotter_path + 'tracer_image_plane_image.png' in plot_patch.paths
    assert ray_tracing_plotter_path + 'tracer_source_plane.png' in plot_patch.paths
    assert ray_tracing_plotter_path + 'tracer_convergence.png' not in plot_patch.paths
    assert ray_tracing_plotter_path + 'tracer_potential.png' in plot_patch.paths
    assert ray_tracing_plotter_path + 'tracer_deflections_y.png' not in plot_patch.paths
    assert ray_tracing_plotter_path + 'tracer_deflections_x.png' not in plot_patch.paths


from autolens.data import convolution
from autolens.model.inversion import convolution as inversion_convolution

class MockConvolverImage(convolution.ConvolverImage):

    def __init__(self, mask, blurring_mask, psf):

        super(MockConvolverImage, self).__init__(mask=mask, blurring_mask=blurring_mask, psf=psf)


class MockConvolverMappingMatrix(inversion_convolution.ConvolverMappingMatrix):

    def __init__(self, mask, psf):

        super(MockConvolverMappingMatrix, self).__init__(mask=mask, psf=psf)


from autolens import decorator_util
import numpy as np

from autolens.data.array.util import array_util, mask_util


@decorator_util.jit()
def padded_binning_shape_2d_from_shape_2d_and_bin_up_factor(shape_2d, bin_up_factor):

    shape_remainder = (shape_2d[0] % bin_up_factor, shape_2d[1] % bin_up_factor)

    if shape_remainder[0] != 0 and shape_remainder[1] != 0:
        shape_pad = (bin_up_factor - shape_remainder[0], bin_up_factor - shape_remainder[1])
    elif shape_remainder[0] != 0 and shape_remainder[1] == 0:
        shape_pad = (bin_up_factor - shape_remainder[0], 0)
    elif shape_remainder[0] == 0 and shape_remainder[1] != 0:
        shape_pad = (0, bin_up_factor - shape_remainder[1])
    else:
        shape_pad = (0, 0)

    return (shape_2d[0] + shape_pad[0], shape_2d[1] + shape_pad[1])

@decorator_util.jit()
def padded_binning_array_2d_from_array_2d_and_bin_up_factor(array_2d, bin_up_factor, pad_value=0.0):
    """If an array is to be binned up, but the dimensions are not divisible by the bin-up factor, this routine pads \
    the array to make it divisible.

    For example, if the array is shape (5,5) and the bin_up_factor is 2, this routine will pad the array to shape \
    (6,6).

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is padded.
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).
    pad_value : float
        If the array is padded, the value the padded edge values are filled in using.

    Returns
    -------
    ndarray
        The 2D array that is padded before binning up.

    Examples
    --------
    array_2d = np.ones((5,5))
    padded_array_2d = padded_array_2d_for_binning_up_with_bin_up_factor( \
        array_2d=array_2d, bin_up_factor=2, pad_value=0.0)
    """

    padded_binning_shape_2d = padded_binning_shape_2d_from_shape_2d_and_bin_up_factor(
        shape_2d=array_2d.shape, bin_up_factor=bin_up_factor)

    return array_util.resized_array_2d_from_array_2d_and_resized_shape(
        array_2d=array_2d, resized_shape=padded_binning_shape_2d, pad_value=pad_value)

@decorator_util.jit()
def binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(array_2d, bin_up_factor):
    """Bin up an array to coarser resolution, by binning up groups of pixels and using their mean value to determine \
     the value of the new pixel.

    If an array of shape (8,8) is input and the bin up size is 2, this would return a new array of size (4,4) where \
    every pixel was the mean of each collection of 2x2 pixels on the (8,8) array.

    If binning up the array leads to an edge being cut (e.g. a (9,9) array binned up by 2), the array is first \
    padded to make the division work. One must be careful of edge effects in this case.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is binned up.
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The binned up 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    resize_array = bin_up_array_2d_using_mean(array_2d=array_2d, bin_up_factor=2)
    """

    padded_binning_array_2d = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=array_2d, bin_up_factor=bin_up_factor)

    binned_array_2d = np.zeros(shape=(padded_binning_array_2d.shape[0] // bin_up_factor,
                                      padded_binning_array_2d.shape[1] // bin_up_factor))

    for y in range(binned_array_2d.shape[0]):
        for x in range(binned_array_2d.shape[1]):
            value = 0.0
            for y1 in range(bin_up_factor):
                for x1 in range(bin_up_factor):
                    padded_y = y*bin_up_factor + y1
                    padded_x = x*bin_up_factor + x1
                    value += padded_binning_array_2d[padded_y, padded_x]

            binned_array_2d[y,x] = value / (bin_up_factor ** 2.0)

    return binned_array_2d

@decorator_util.jit()
def binned_array_2d_using_quadrature_from_array_2d_and_bin_up_factor(array_2d, bin_up_factor):
    """Bin up an array to coarser resolution, by binning up groups of pixels and using their quadrature value to \
    determine the value of the new pixel.

    If an array of shape (8,8) is input and the bin up size is 2, this would return a new array of size (4,4) where \
    every pixel was the quadrature of each collection of 2x2 pixels on the (8,8) array.

    If binning up the array leads to an edge being cut (e.g. a (9,9) array binned up by 2), the array is first \
    padded to make the division work. One must be careful of edge effects in this case.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is binned up.
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The binned up 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    resize_array = bin_up_array_2d_using_quadrature(array_2d=array_2d, bin_up_factor=2)
    """

    padded_binning_array_2d = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=array_2d, bin_up_factor=bin_up_factor)

    binned_array_2d = np.zeros(shape=(padded_binning_array_2d.shape[0] // bin_up_factor,
                                      padded_binning_array_2d.shape[1] // bin_up_factor))

    for y in range(binned_array_2d.shape[0]):
        for x in range(binned_array_2d.shape[1]):
            value = 0.0
            for y1 in range(bin_up_factor):
                for x1 in range(bin_up_factor):
                    padded_y = y*bin_up_factor + y1
                    padded_x = x*bin_up_factor + x1
                    value += padded_binning_array_2d[padded_y, padded_x] ** 2.0

            binned_array_2d[y,x] = np.sqrt(value) / (bin_up_factor ** 2.0)

    return binned_array_2d

@decorator_util.jit()
def binned_array_2d_using_sum_from_array_2d_and_bin_up_factor(array_2d, bin_up_factor):
    """Bin up an array to coarser resolution, by binning up groups of pixels and using their sum value to determine \
     the value of the new pixel.

    If an array of shape (8,8) is input and the bin up size is 2, this would return a new array of size (4,4) where \
    every pixel was the sum of each collection of 2x2 pixels on the (8,8) array.

    If binning up the array leads to an edge being cut (e.g. a (9,9) array binned up by 2), the array is first \
    padded to make the division work. One must be careful of edge effects in this case.

    Parameters
    ----------
    array_2d : ndarray
        The 2D array that is binned up.
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The binned up 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    resize_array = bin_up_array_2d_using_sum(array_2d=array_2d, bin_up_factor=2)
    """

    padded_binning_array_2d = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=array_2d, bin_up_factor=bin_up_factor)

    binned_array_2d = np.zeros(shape=(padded_binning_array_2d.shape[0] // bin_up_factor,
                                      padded_binning_array_2d.shape[1] // bin_up_factor))

    for y in range(binned_array_2d.shape[0]):
        for x in range(binned_array_2d.shape[1]):
            value = 0.0
            for y1 in range(bin_up_factor):
                for x1 in range(bin_up_factor):
                    padded_y = y*bin_up_factor + y1
                    padded_x = x*bin_up_factor + x1
                    value += padded_binning_array_2d[padded_y, padded_x]

            binned_array_2d[y,x] = value

    return binned_array_2d

@decorator_util.jit()
def binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d, bin_up_factor):
    """Bin up an array to coarser resolution, by binning up groups of pixels and using their sum value to determine \
     the value of the new pixel.

    If an array of shape (8,8) is input and the bin up size is 2, this would return a new array of size (4,4) where \
    every pixel was the sum of each collection of 2x2 pixels on the (8,8) array.

    If binning up the array leads to an edge being cut (e.g. a (9,9) array binned up by 2), an array is first \
    extracted around the centre of that array.


    Parameters
    ----------
    mask_2d : ndarray
        The 2D array that is resized.
    new_shape : (int, int)
        The (y,x) new pixel dimension of the trimmed array.
    origin : (int, int)
        The oigin of the resized array, e.g. the central pixel around which the array is extracted.

    Returns
    -------
    ndarray
        The resized 2D array from the input 2D array.

    Examples
    --------
    array_2d = np.ones((5,5))
    resize_array = resize_array_2d(array_2d=array_2d, new_shape=(2,2), origin=(2, 2))
    """

    padded_mask_2d = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=mask_2d, bin_up_factor=bin_up_factor, pad_value=True)

    binned_mask_2d = np.zeros(shape=(padded_mask_2d.shape[0] // bin_up_factor,
                                      padded_mask_2d.shape[1] // bin_up_factor))

    for y in range(binned_mask_2d.shape[0]):
        for x in range(binned_mask_2d.shape[1]):
            value = True
            for y1 in range(bin_up_factor):
                for x1 in range(bin_up_factor):
                    padded_y = y*bin_up_factor + y1
                    padded_x = x*bin_up_factor + x1
                    if padded_mask_2d[padded_y, padded_x] == False:
                        value = False

            binned_mask_2d[y,x] = value

    return binned_mask_2d

@decorator_util.jit()
def binned_mask_2d_mapper_from_mask_2d_and_bin_up_factor(mask_2d, bin_up_factor):
    """Create a 2D mask which effectively maps a 2D mask to a 2D binned mask. wE create a mask the same shape \
    as the 2D mask, but in each binned up region set all values equal to True except for one value (the central \
    value).
    
    This provides us with a mask which we can iterate over to count the indexes of binned up pixels, whilst using \ 
    the original 2D mask to keep track of the mask's indexes. This is used for mapping binned up masked array index's \
    to the original mask's indexes.
    
    For example, if we had a 6x6 mask of all False:
    
    [[False, False, False, False, False, False],
     [False, False, False, False, False, False],
     [False, False, False, False, False, False],
     [False, False, False, False, False, False],
     [False, False, False, False, False, False],
     [False, False, False, False, False, False]]
     
    For a bin_up_factor of 3, the resulting padded mask mapper is as follows:
    
    [[True, True, True, True, True, True],
     [True, False, True, True, False, True],
     [True, True, True, True, True, True],
     [True, True, True, True, True, True],
     [True, False, True, True, False, True],
     [True, True, True, True, True, True]]

    The above mask then masks it much easier to pair masked binned up indexes to the original mask's indexes.

    Parameters
    ----------
    mask_2d : ndarray
        The 2D mask that the
    bin_up_factor : int
        The factor which the array is binned up by (e.g. a value of 2 bins every 2 x 2 pixels into one pixel).

    Returns
    -------
    ndarray
        The binned up 2D array from the input 2D array.

    Examples
    --------
    mask_2d = np.full(fill_value=False, shape=(9,9))
    binned_mask_2d_mapper = binned_mask_2d_mapper_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=3)
    """

    padded_mask_2d = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=mask_2d, bin_up_factor=bin_up_factor, pad_value=True)

    binned_mask_2d = binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=bin_up_factor)

    binned_mask_2d_mapper = np.full(fill_value=True, shape=padded_mask_2d.shape)

    for bin_y in range(binned_mask_2d.shape[0]):
        for bin_x in range(binned_mask_2d.shape[1]):
            if binned_mask_2d[bin_y, bin_x] == False:
                mask_entry_found = False
                for bin_y1 in range(bin_up_factor):
                    for bin_x1 in range(bin_up_factor):
                        mask_y = bin_y*bin_up_factor + bin_y1
                        mask_x = bin_x*bin_up_factor + bin_x1
                        if padded_mask_2d[mask_y, mask_x] == False and not mask_entry_found:
                            binned_mask_2d_mapper[mask_y, mask_x] = False
                            mask_entry_found = True

    return binned_mask_2d_mapper

@decorator_util.jit()
def binned_masked_array_1d_to_masked_array_1d_from_mask_2d_and_bin_up_factor(mask_2d, bin_up_factor):

    padded_mask_2d = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=mask_2d, bin_up_factor=bin_up_factor, pad_value=True)

    binned_mask_2d_mapper = binned_mask_2d_mapper_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=bin_up_factor)

    total_binned_masked_pixels = mask_util.total_regular_pixels_from_mask(mask=binned_mask_2d_mapper)

    binned_masked_array_1d_to_masked_array_1d = np.zeros(total_binned_masked_pixels)

    binned_mask_index = 0
    mask_index = 0

    for mask_y in range(padded_mask_2d.shape[0]):
        for mask_x in range(padded_mask_2d.shape[1]):
            if binned_mask_2d_mapper[mask_y, mask_x] == False:
                binned_masked_array_1d_to_masked_array_1d[binned_mask_index] = mask_index
                binned_mask_index += 1
            if padded_mask_2d[mask_y, mask_x] == False:
                mask_index += 1

    return binned_masked_array_1d_to_masked_array_1d

@decorator_util.jit()
def masked_array_1d_to_binned_masked_array_1d_from_mask_2d_and_bin_up_factor(mask_2d, bin_up_factor):

    padded_mask_2d = padded_binning_array_2d_from_array_2d_and_bin_up_factor(
        array_2d=mask_2d, bin_up_factor=bin_up_factor, pad_value=True)

    binned_mask_2d = binned_up_mask_2d_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=bin_up_factor)

    binned_mask_2d_mapper = binned_mask_2d_mapper_from_mask_2d_and_bin_up_factor(mask_2d=mask_2d, bin_up_factor=bin_up_factor)

    total_masked_pixels = mask_util.total_regular_pixels_from_mask(mask=padded_mask_2d)

    masked_array_1d_to_binned_masked_array_1d = np.zeros(total_masked_pixels)

    binned_mask_index = 0
    mask_index = 0

    for bin_y in range(binned_mask_2d.shape[0]):
        for bin_x in range(binned_mask_2d.shape[1]):
            if binned_mask_2d[bin_y, bin_x] == False:
                for bin_y1 in range(bin_up_factor):
                    for bin_x1 in range(bin_up_factor):
                        mask_y = bin_y*bin_up_factor + bin_y1
                        mask_x = bin_x*bin_up_factor + bin_x1
                        if padded_mask_2d[mask_y, mask_x] == False and not mask_entry_found:
                            binned_mask_2d_mapper[mask_y, mask_x] = False
                            mask_entry_found = True

    for mask_y in range(padded_mask_2d.shape[0]):
        for mask_x in range(padded_mask_2d.shape[1]):
            if binned_mask_2d_mapper[mask_y, mask_x] == False:
                masked_array_1d_to_binned_masked_array_1d[binned_mask_index] = mask_index
                binned_mask_index += 1
            if padded_mask_2d[mask_y, mask_x] == False:
                mask_index += 1

    return masked_array_1d_to_binned_masked_array_1d


from autolens.lens.plotters import lens_plotter_util
from test.fixtures import *


@pytest.fixture(name='lens_plotter_util_path')
def make_lens_plotter_util_path_setup():
    return "{}/../../test_files/plotting/lens_plotter_util/".format(
        os.path.dirname(os.path.realpath(__file__)))

def test__image_is_output(lens_fit_x2_plane_5x5, lens_plotter_util_path, plot_patch):

    lens_plotter_util.plot_image(fit=lens_fit_x2_plane_5x5, mask=lens_fit_x2_plane_5x5.mask_2d, extract_array_from_mask=True,
                                 zoom_around_mask=True, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                 output_path=lens_plotter_util_path, output_format='png')

    assert lens_plotter_util_path + 'fit_image.png' in plot_patch.paths


def test__noise_map_is_output(lens_fit_x2_plane_5x5, lens_plotter_util_path, plot_patch):

    lens_plotter_util.plot_noise_map(
        fit=lens_fit_x2_plane_5x5, mask=lens_fit_x2_plane_5x5.mask_2d, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'], output_path=lens_plotter_util_path, output_format='png')

    assert lens_plotter_util_path + 'fit_noise_map.png' in plot_patch.paths


def test__signal_to_noise_map_is_output(lens_fit_x2_plane_5x5, lens_plotter_util_path, plot_patch):
    lens_plotter_util.plot_signal_to_noise_map(
        fit=lens_fit_x2_plane_5x5, mask=lens_fit_x2_plane_5x5.mask_2d, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'], output_path=lens_plotter_util_path, output_format='png')

    assert lens_plotter_util_path + 'fit_signal_to_noise_map.png' in plot_patch.paths


def test__model_image_is_output(lens_fit_x2_plane_5x5, lens_plotter_util_path, plot_patch):

    lens_plotter_util.plot_model_data(
        fit=lens_fit_x2_plane_5x5, mask=lens_fit_x2_plane_5x5.mask_2d, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'], output_path=lens_plotter_util_path, output_format='png')

    assert lens_plotter_util_path + 'fit_model_image.png' in plot_patch.paths


def test__residual_map_is_output(lens_fit_x2_plane_5x5, lens_plotter_util_path, plot_patch):
    lens_plotter_util.plot_residual_map(
        fit=lens_fit_x2_plane_5x5, mask=lens_fit_x2_plane_5x5.mask_2d, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'], output_path=lens_plotter_util_path, output_format='png')

    assert lens_plotter_util_path + 'fit_residual_map.png' in plot_patch.paths


def test__chi_squared_map_is_output(lens_fit_x2_plane_5x5, lens_plotter_util_path, plot_patch):

    lens_plotter_util.plot_chi_squared_map(
        fit=lens_fit_x2_plane_5x5, mask=lens_fit_x2_plane_5x5.mask_2d, extract_array_from_mask=True, zoom_around_mask=True,
        cb_tick_values=[1.0], cb_tick_labels=['1.0'], output_path=lens_plotter_util_path, output_format='png')

    assert lens_plotter_util_path + 'fit_chi_squared_map.png' in plot_patch.paths



import numpy as np
from astropy import cosmology as cosmo

import autofit as af
from autolens.data.array.util import binning_util
from autolens import exc
from autolens.data.array import grids
from autolens.data.plotters import ccd_plotters
from autolens.lens import ray_tracing, lens_data as ld, lens_fit, sensitivity_fit
from autolens.lens.plotters import ray_tracing_plotters, lens_fit_plotters, \
    sensitivity_fit_plotters
from autolens.model.galaxy import galaxy as g
from autolens.model.inversion import pixelizations as px
from autolens.pipeline import tagging as tag
from autolens.pipeline.phase.phase import Phase, setup_phase_mask


class PhaseImaging(Phase):

    def __init__(self, phase_name, tag_phases=True, phase_folders=None,
                 optimizer_class=af.MultiNest,
                 sub_grid_size=2, bin_up_factor=None, image_psf_shape=None,
                 inversion_psf_shape=None, positions_threshold=None, mask_function=None,
                 inner_mask_radii=None,
                 interp_pixel_scale=None,
                 cluster_pixel_scale=None,
                 cosmology=cosmo.Planck15,
                 auto_link_priors=False):

        """

        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper
        passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """

        if tag_phases:

            phase_tag = tag.phase_tag_from_phase_settings(sub_grid_size=sub_grid_size,
                                                          bin_up_factor=bin_up_factor,
                                                          image_psf_shape=image_psf_shape,
                                                          inversion_psf_shape=inversion_psf_shape,
                                                          positions_threshold=positions_threshold,
                                                          inner_mask_radii=inner_mask_radii,
                                                          interp_pixel_scale=interp_pixel_scale)

        else:

            phase_tag = None

        super(PhaseImaging, self).__init__(phase_name=phase_name, phase_tag=phase_tag,
                                           phase_folders=phase_folders,
                                           tag_phases=tag_phases,
                                           optimizer_class=optimizer_class,
                                           cosmology=cosmology,
                                           auto_link_priors=auto_link_priors)

        self.sub_grid_size = sub_grid_size
        self.bin_up_factor = bin_up_factor
        self.image_psf_shape = image_psf_shape
        self.inversion_psf_shape = inversion_psf_shape
        self.positions_threshold = positions_threshold
        self.mask_function = mask_function
        self.inner_mask_radii = inner_mask_radii
        self.interp_pixel_scale = interp_pixel_scale
        self.cluster_pixel_scale = cluster_pixel_scale


    @property
    def uses_hyper_images(self) -> bool:
        return False

    @property
    def uses_inversion(self) -> bool:
        return False

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def modify_image(self, image, results):
        """
        Customize an lens_data. e.g. removing lens light.

        Parameters
        ----------
        image: scaled_array.ScaledSquarePixelArray
            An lens_data that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result of the previous lens

        Returns
        -------
        lens_data: scaled_array.ScaledSquarePixelArray
            The modified image (not changed by default)
        """
        return image

    def run(self, data, results=None, mask=None, positions=None):
        """
        Run this phase.

        Parameters
        ----------
        positions
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed
        data: scaled_array.ScaledSquarePixelArray
            An lens_data that has been masked

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper.
        """
        analysis = self.make_analysis(data=data, results=results, mask=mask,
                                      positions=positions)

        self.pass_priors(results)
        self.assert_and_save_pickle()

        result = self.run_analysis(analysis)

        return self.make_result(result, analysis)

    def make_analysis(self, data, results=None, mask=None, positions=None):
        """
        Create an lens object. Also calls the prior passing and lens_data modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        positions
        mask: Mask
            The default masks passed in by the pipeline
        data: im.CCD
            An lens_data that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens : Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a set of values
        """

        mask = setup_phase_mask(data=data, mask=mask, mask_function=self.mask_function,
                                inner_mask_radii=self.inner_mask_radii)

        if self.positions_threshold is not None and positions is not None:
            positions = list(map(lambda position_set: np.asarray(position_set), positions))
        elif self.positions_threshold is None:
            positions = None
        elif self.positions_threshold is not None and positions is None:
            raise exc.PhaseException(
                'You have specified for a phase to use positions, but not input positions to the '
                'pipeline when you ran it.')

        lens_data = ld.LensData(ccd_data=data, mask=mask,
                                sub_grid_size=self.sub_grid_size,
                                image_psf_shape=self.image_psf_shape,
                                positions=positions,
                                interp_pixel_scale=self.interp_pixel_scale,
                                cluster_pixel_scale=self.cluster_pixel_scale,
                                uses_inversion=self.uses_inversion)

        modified_image = self.modify_image(image=lens_data.unmasked_image,
                                           results=results)
        lens_data = lens_data.new_lens_data_with_modified_image(
            modified_image=modified_image)

        if self.bin_up_factor is not None:

            lens_data = lens_data.new_lens_data_with_binned_up_ccd_data_and_mask(
                bin_up_factor=self.bin_up_factor)

        self.output_phase_info()

        analysis = self.__class__.Analysis(lens_data=lens_data,
                                           cosmology=self.cosmology,
                                           positions_threshold=self.positions_threshold,
                                           results=results,
                                           uses_hyper_images=self.uses_hyper_images)
        return analysis

    def output_phase_info(self):

        file_phase_info = "{}/{}".format(self.optimizer.phase_output_path, 'phase.info')

        with open(file_phase_info, 'w') as phase_info:
            phase_info.write('Optimizer = {} \n'.format(type(self.optimizer).__name__))
            phase_info.write('Sub-grid size = {} \n'.format(self.sub_grid_size))
            phase_info.write('Image PSF shape = {} \n'.format(self.image_psf_shape))
            phase_info.write(
                'Pixelization PSF shape = {} \n'.format(self.inversion_psf_shape))
            phase_info.write(
                'Positions Threshold = {} \n'.format(self.positions_threshold))
            phase_info.write('Cosmology = {} \n'.format(self.cosmology))
            phase_info.write('Auto Link Priors = {} \n'.format(self.auto_link_priors))

            phase_info.close()

    # noinspection PyAbstractClass
    class Analysis(Phase.Analysis):

        def __init__(self, lens_data, cosmology, positions_threshold, results=None,
                     uses_hyper_images=False):

            super(PhaseImaging.Analysis, self).__init__(cosmology=cosmology,
                                                        results=results)

            self.lens_data = lens_data

            self.positions_threshold = positions_threshold

            self.should_plot_image_plane_pix = \
                af.conf.instance.general.get('output',
                                             'plot_image_plane_adaptive_pixelization_grid',
                                             bool)

            self.plot_data_as_subplot = \
                af.conf.instance.general.get('output', 'plot_data_as_subplot', bool)
            self.plot_data_image = \
                af.conf.instance.general.get('output', 'plot_data_image', bool)
            self.plot_data_noise_map = \
                af.conf.instance.general.get('output', 'plot_data_noise_map', bool)
            self.plot_data_psf = \
                af.conf.instance.general.get('output', 'plot_data_psf', bool)
            self.plot_data_signal_to_noise_map = \
                af.conf.instance.general.get('output', 'plot_data_signal_to_noise_map',
                                             bool)
            self.plot_data_absolute_signal_to_noise_map = \
                af.conf.instance.general.get('output',
                                             'plot_data_absolute_signal_to_noise_map',
                                             bool)
            self.plot_data_potential_chi_squared_map = \
                af.conf.instance.general.get('output',
                                             'plot_data_potential_chi_squared_map',
                                             bool)

            self.plot_lens_fit_all_at_end_png = \
                af.conf.instance.general.get('output', 'plot_lens_fit_all_at_end_png',
                                             bool)
            self.plot_lens_fit_all_at_end_fits = \
                af.conf.instance.general.get('output', 'plot_lens_fit_all_at_end_fits',
                                             bool)

            self.plot_lens_fit_as_subplot = \
                af.conf.instance.general.get('output', 'plot_lens_fit_as_subplot', bool)
            self.plot_lens_fit_image = \
                af.conf.instance.general.get('output', 'plot_lens_fit_image', bool)
            self.plot_lens_fit_noise_map = \
                af.conf.instance.general.get('output', 'plot_lens_fit_noise_map', bool)
            self.plot_lens_fit_signal_to_noise_map = \
                af.conf.instance.general.get('output',
                                             'plot_lens_fit_signal_to_noise_map',
                                             bool)
            self.plot_lens_fit_lens_subtracted_image = \
                af.conf.instance.general.get('output',
                                             'plot_lens_fit_lens_subtracted_image',
                                             bool)
            self.plot_lens_fit_model_image = \
                af.conf.instance.general.get('output', 'plot_lens_fit_model_image',
                                             bool)
            self.plot_lens_fit_lens_model_image = \
                af.conf.instance.general.get('output', 'plot_lens_fit_lens_model_image',
                                             bool)
            self.plot_lens_fit_source_model_image = \
                af.conf.instance.general.get('output',
                                             'plot_lens_fit_source_model_image',
                                             bool)
            self.plot_lens_fit_source_plane_image = \
                af.conf.instance.general.get('output',
                                             'plot_lens_fit_source_plane_image',
                                             bool)
            self.plot_lens_fit_residual_map = \
                af.conf.instance.general.get('output', 'plot_lens_fit_residual_map',
                                             bool)
            self.plot_lens_fit_chi_squared_map = \
                af.conf.instance.general.get('output', 'plot_lens_fit_chi_squared_map',
                                             bool)

            self.plot_lens_fit_contribution_maps = \
                af.conf.instance.general.get('output',
                                             'plot_lens_fit_contribution_maps',
                                             bool)
            self.plot_lens_fit_regularization_weights = \
                af.conf.instance.general.get('output',
                                             'plot_lens_fit_regularization_weights',
                                             bool)

            self.uses_hyper_images = uses_hyper_images

            if self.last_results is not None and self.uses_hyper_images:

                image_1d_galaxy_dict = {}

                self.hyper_model_image_1d = np.zeros(lens_data.mask_1d.shape)

                for galaxy, galaxy_image in self.last_results.image_2d_dict.items():

                    image_1d_galaxy_dict[galaxy] = lens_data.array_1d_from_array_2d(array_2d=galaxy_image)
                    self.check_for_previously_masked_values(array=image_1d_galaxy_dict[galaxy])

                self.hyper_galaxy_image_1d_path_dict = {}

                for path, galaxy in self.last_results.path_galaxy_tuples:

                    galaxy_image = image_1d_galaxy_dict[path]

                    self.hyper_model_image_1d += galaxy_image

                    minimum_galaxy_value = 0.01*max(galaxy_image)
                    galaxy_image[galaxy_image < minimum_galaxy_value] = minimum_galaxy_value

                    self.hyper_galaxy_image_1d_path_dict[path] = galaxy_image

                cluster_image_1d_galaxy_dict = {}

                for galaxy, galaxy_image in self.last_results.image_2d_dict.items():

                    cluster_image_2d = binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
                        array_2d=galaxy_image, bin_up_factor=lens_data.cluster.bin_up_factor)

                    cluster_image_1d_galaxy_dict[galaxy] = \
                        lens_data.cluster.mask.map_2d_array_to_masked_1d_array(array_2d=cluster_image_2d)

                self.hyper_galaxy_cluster_image_1d_path_dict = {}

                for path, galaxy in self.last_results.path_galaxy_tuples:

                    galaxy_cluster_image = cluster_image_1d_galaxy_dict[path]

                    minimum_cluster_value = 0.01 * max(galaxy_cluster_image)
                    galaxy_cluster_image[galaxy_cluster_image < minimum_cluster_value] = minimum_cluster_value

                    self.hyper_galaxy_cluster_image_1d_path_dict[path] = galaxy_cluster_image


            else:

                self.hyper_galaxy_image_1d_path_dict = None
                self.hyper_model_image_1d = None

        def fit(self, instance):
            """
            Determine the fit of a lens galaxy and source galaxy to the lens_data in this lens.

            Parameters
            ----------
            instance
                A model instance with attributes

            Returns
            -------
            fit : Fit
                A fractional value indicating how well this model fit and the model lens_data itself
            """
            self.check_positions_trace_within_threshold(instance=instance)
            tracer = self.tracer_for_instance(instance=instance)
            fit = self.fit_for_tracers(tracer=tracer, padded_tracer=None)
            return fit.figure_of_merit

        def check_for_previously_masked_values(self, array):
            if not np.all(array) != 0.0 and not np.all(array == 0):
                raise exc.PhaseException(
                    'When mapping a 2D array to a 1D array using lens data, a value '
                    'encountered was 0.0 and therefore masked in a previous phase.')

        def associate_images(self, instance: af.ModelInstance) -> af.ModelInstance:
            """
            Takes images from the last result, if there is one, and associates them with galaxies in this phase where
            full-path galaxy names match.

            If the galaxy collection has a different name then an association is not made.

            e.g.
            lens_galaxies.lens will match with:
                lens_galaxies.lens
            but not with:
                galaxies.lens
                lens_galaxies.source

            Parameters
            ----------
            instance
                A model instance with 0 or more galaxies in its tree

            Returns
            -------
            instance
               The input instance with images associated with galaxies where possible.
            """
            if self.uses_hyper_images:
                for galaxy_path, galaxy in instance.path_instance_tuples_for_class(g.Galaxy):
                    if galaxy_path in self.hyper_galaxy_image_1d_path_dict:
                        galaxy.hyper_model_image_1d = self.hyper_model_image_1d
                        galaxy.hyper_galaxy_image_1d = self.hyper_galaxy_image_1d_path_dict[galaxy_path]
                        galaxy.hyper_galaxy_cluster_image_1d = self.hyper_galaxy_cluster_image_1d_path_dict[galaxy_path]
            return instance

        def add_grids_to_grid_stack(self, galaxies, grid_stack):

            for galaxy in galaxies:
                if galaxy.pixelization is not None:
                    if galaxy.pixelization.uses_pixelization_grid:

                        if isinstance(galaxy.pixelization, px.VoronoiMagnification):

                            sparse_to_regular_grid = grids.SparseToRegularGrid.from_unmasked_2d_grid_shape_and_regular_grid(
                                unmasked_sparse_shape=galaxy.pixelization.shape, regular_grid=grid_stack.regular)

                        elif isinstance(galaxy.pixelization, px.VoronoiBrightnessImage):

                            cluster_weight_map = galaxy.pixelization.cluster_weight_map_from_hyper_image(
                                    hyper_image=galaxy.hyper_galaxy_cluster_image_1d)

                            sparse_to_regular_grid = \
                                grids.SparseToRegularGrid.from_total_pixels_cluster_grid_and_cluster_weight_map(
                                    total_pixels=galaxy.pixelization.pixels, cluster_grid=self.lens_data.cluster,
                                    regular_grid=self.lens_data.grid_stack.regular, cluster_weight_map=cluster_weight_map, seed=1)

                        else:

                            raise exc.PhaseException('The pixelization of a galaxy uses a pixelization grid, but was not a viable'
                                                     'type in the grid stack calculation method')

                        pixelization_grid = grids.PixelizationGrid(
                            arr=sparse_to_regular_grid.sparse, regular_to_pixelization=sparse_to_regular_grid.regular_to_sparse)

                        return grid_stack.new_grid_stack_with_grids_added(pixelization=pixelization_grid)

            return grid_stack

        def visualize(self, instance, image_path, during_analysis):

            instance = self.associate_images(instance=instance)

            mask = self.lens_data.mask_2d if self.should_plot_mask else None
            positions = self.lens_data.positions if self.should_plot_positions else None

            ccd_plotters.plot_ccd_for_phase(
                ccd_data=self.lens_data.ccd_data, mask=mask, positions=positions,
                extract_array_from_mask=self.extract_array_from_mask,
                zoom_around_mask=self.zoom_around_mask,
                units=self.plot_units,
                should_plot_as_subplot=self.plot_data_as_subplot,
                should_plot_image=self.plot_data_image,
                should_plot_noise_map=self.plot_data_noise_map,
                should_plot_psf=self.plot_data_psf,
                should_plot_signal_to_noise_map=self.plot_data_signal_to_noise_map,
                should_plot_absolute_signal_to_noise_map=self.plot_data_absolute_signal_to_noise_map,
                should_plot_potential_chi_squared_map=self.plot_data_potential_chi_squared_map,
                visualize_path=image_path)

            tracer = self.tracer_for_instance(instance)

            ray_tracing_plotters.plot_ray_tracing_for_phase(
                tracer=tracer, during_analysis=during_analysis, mask=mask,
                extract_array_from_mask=self.extract_array_from_mask,
                zoom_around_mask=self.zoom_around_mask, positions=positions,
                units=self.plot_units,
                should_plot_as_subplot=self.plot_ray_tracing_as_subplot,
                should_plot_all_at_end_png=self.plot_ray_tracing_all_at_end_png,
                should_plot_all_at_end_fits=self.plot_ray_tracing_all_at_end_fits,
                should_plot_image_plane_image=self.plot_ray_tracing_image_plane_image,
                should_plot_source_plane=self.plot_ray_tracing_source_plane,
                should_plot_convergence=self.plot_ray_tracing_convergence,
                should_plot_potential=self.plot_ray_tracing_potential,
                should_plot_deflections=self.plot_ray_tracing_deflections,
                visualize_path=image_path)

            padded_tracer = self.padded_tracer_for_instance(instance)
            fit = self.fit_for_tracers(tracer=tracer, padded_tracer=padded_tracer)

            lens_fit_plotters.plot_lens_fit_for_phase(
                fit=fit, during_analysis=during_analysis,
                should_plot_mask=self.should_plot_mask,
                extract_array_from_mask=self.extract_array_from_mask,
                zoom_around_mask=self.zoom_around_mask,
                positions=positions,
                should_plot_image_plane_pix=self.should_plot_image_plane_pix,
                should_plot_as_subplot=self.plot_lens_fit_as_subplot,
                should_plot_all_at_end_png=self.plot_lens_fit_all_at_end_png,
                should_plot_all_at_end_fits=self.plot_lens_fit_all_at_end_fits,
                should_plot_image=self.plot_lens_fit_image,
                should_plot_noise_map=self.plot_lens_fit_noise_map,
                should_plot_signal_to_noise_map=self.plot_lens_fit_signal_to_noise_map,
                should_plot_lens_subtracted_image=self.plot_lens_fit_lens_subtracted_image,
                should_plot_model_image=self.plot_lens_fit_model_image,
                should_plot_lens_model_image=self.plot_lens_fit_lens_model_image,
                should_plot_source_model_image=self.plot_lens_fit_source_model_image,
                should_plot_source_plane_image=self.plot_lens_fit_source_plane_image,
                should_plot_residual_map=self.plot_lens_fit_residual_map,
                should_plot_chi_squared_map=self.plot_lens_fit_chi_squared_map,
                should_plot_regularization_weights=self.plot_lens_fit_regularization_weights,
                units=self.plot_units,
                visualize_path=image_path)

        def fit_for_tracers(self, tracer, padded_tracer):
            return lens_fit.LensDataFit.for_data_and_tracer(lens_data=self.lens_data,
                                                            tracer=tracer,
                                                            padded_tracer=padded_tracer)

        def check_positions_trace_within_threshold(self, instance):

            if self.lens_data.positions is not None:

                tracer = ray_tracing.TracerImageSourcePlanesPositions(
                    lens_galaxies=instance.lens_galaxies,
                    image_plane_positions=self.lens_data.positions)
                fit = lens_fit.LensPositionFit(positions=tracer.source_plane.positions,
                                               noise_map=self.lens_data.pixel_scale)

                if not fit.maximum_separation_within_threshold(
                        self.positions_threshold):
                    raise exc.RayTracingException

        def map_to_1d(self, data):
            """Convenience method"""
            return self.lens_data.mask.map_2d_array_to_masked_1d_array(data)


class MultiPlanePhase(PhaseImaging):
    """
    Fit a simple source and lens system.
    """

    galaxies = af.PhaseProperty("galaxies")

    def __init__(self, phase_name, tag_phases=True, phase_folders=None, galaxies=None,
                 optimizer_class=af.MultiNest,
                 sub_grid_size=2, bin_up_factor=None, image_psf_shape=None,
                 positions_threshold=None,
                 mask_function=None,
                 inner_mask_radii=None,
                 cluster_pixel_scale=None,
                 cosmology=cosmo.Planck15,
                 auto_link_priors=False):
        """
        A phase with a simple source/lens model

        Parameters
        ----------
        galaxies : [g.Galaxy] | [gm.GalaxyModel]
            A galaxy that acts as a gravitational lens or is being lensed
        optimizer_class: class
            The class of a non-linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """

        super(MultiPlanePhase, self).__init__(phase_name=phase_name,
                                              tag_phases=tag_phases,
                                              phase_folders=phase_folders,
                                              optimizer_class=optimizer_class,
                                              sub_grid_size=sub_grid_size,
                                              bin_up_factor=bin_up_factor,
                                              image_psf_shape=image_psf_shape,
                                              positions_threshold=positions_threshold,
                                              mask_function=mask_function,
                                              inner_mask_radii=inner_mask_radii,
                                              cluster_pixel_scale=cluster_pixel_scale,
                                              cosmology=cosmology,
                                              auto_link_priors=auto_link_priors)
        self.galaxies = galaxies

    @property
    def uses_hyper_images(self):
        if self.galaxies:
            return any([galaxy.uses_hyper_images for galaxy in self.galaxies])
        else:
            return False

    @property
    def uses_inversion(self):
        if self.galaxies:
            for galaxy in self.galaxies:
                if galaxy.pixelization is not None:
                    return True
        return False

    class Analysis(PhaseImaging.Analysis):

        def figure_of_merit_for_fit(self, tracer):
            raise NotImplementedError()

        def tracer_for_instance(self, instance):

            instance = self.associate_images(instance=instance)

            image_plane_grid_stack = self.add_grids_to_grid_stack(
                galaxies=instance.galaxies, grid_stack=self.lens_data.grid_stack)

            return ray_tracing.TracerMultiPlanes(galaxies=instance.galaxies,
                                                 image_plane_grid_stack=image_plane_grid_stack,
                                                 border=self.lens_data.border,
                                                 cosmology=self.cosmology)

        def padded_tracer_for_instance(self, instance):

            instance = self.associate_images(instance=instance)

            return ray_tracing.TracerMultiPlanes(galaxies=instance.galaxies,
                                                 image_plane_grid_stack=self.lens_data.padded_grid_stack,
                                                 cosmology=self.cosmology)

        @classmethod
        def describe(cls, instance):
            return "\nRunning multi-plane for... \n\nGalaxies:\n{}\n\n".format(
                instance.galaxies)


class LensSourcePlanePhase(PhaseImaging):
    """
    Fit a simple source and lens system.
    """

    lens_galaxies = af.PhaseProperty("lens_galaxies")
    source_galaxies = af.PhaseProperty("source_galaxies")

    def __init__(self, phase_name, tag_phases=True, phase_folders=None,
                 lens_galaxies=None, source_galaxies=None,
                 optimizer_class=af.MultiNest,
                 sub_grid_size=2, bin_up_factor=None, image_psf_shape=None,
                 positions_threshold=None,
                 mask_function=None,
                 interp_pixel_scale=None, inner_mask_radii=None,
                 cluster_pixel_scale=None,
                 cosmology=cosmo.Planck15,
                 auto_link_priors=False):
        """
        A phase with a simple source/lens model

        Parameters
        ----------
        lens_galaxies : [g.Galaxy] | [gm.GalaxyModel]
            A galaxy that acts as a gravitational lens
        source_galaxies: [g.Galaxy] | [gm.GalaxyModel]
            A galaxy that is being lensed
        optimizer_class: class
            The class of a non-linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """
        super(LensSourcePlanePhase, self).__init__(phase_name=phase_name,
                                                   tag_phases=tag_phases,
                                                   phase_folders=phase_folders,
                                                   optimizer_class=optimizer_class,
                                                   sub_grid_size=sub_grid_size,
                                                   bin_up_factor=bin_up_factor,
                                                   image_psf_shape=image_psf_shape,
                                                   positions_threshold=positions_threshold,
                                                   mask_function=mask_function,
                                                   interp_pixel_scale=interp_pixel_scale,
                                                   inner_mask_radii=inner_mask_radii,
                                                   cluster_pixel_scale=cluster_pixel_scale,
                                                   cosmology=cosmology,
                                                   auto_link_priors=auto_link_priors)
        self.lens_galaxies = lens_galaxies or []
        self.source_galaxies = source_galaxies or []

    @property
    def uses_inversion(self):

        if self.lens_galaxies:
            for galaxy_model in self.lens_galaxies:
                if galaxy_model.pixelization is not None:
                    return True

        if self.source_galaxies:
            for galaxy_model in self.source_galaxies:
                if galaxy_model.pixelization is not None:
                    return True
        return False

    @property
    def uses_hyper_images(self):
        return any([galaxy.uses_hyper_images for galaxy in self.lens_galaxies + self.source_galaxies])

    class Analysis(PhaseImaging.Analysis):

        def figure_of_merit_for_fit(self, tracer):
            raise NotImplementedError()

        def tracer_for_instance(self, instance):

            instance = self.associate_images(instance=instance)

            image_plane_grid_stack = self.add_grids_to_grid_stack(
                galaxies=instance.source_galaxies, grid_stack=self.lens_data.grid_stack)

            return ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=instance.lens_galaxies,
                source_galaxies=instance.source_galaxies,
                image_plane_grid_stack=image_plane_grid_stack,
                border=self.lens_data.border, cosmology=self.cosmology)

        def padded_tracer_for_instance(self, instance):

            instance = self.associate_images(instance=instance)

            return ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=instance.lens_galaxies,
                source_galaxies=instance.source_galaxies,
                image_plane_grid_stack=self.lens_data.padded_grid_stack,
                cosmology=self.cosmology)

        @classmethod
        def describe(cls, instance):
            return "\nRunning lens/source lens for... \n\nLens Galaxy:\n{}\n\nSource " \
                   "Galaxy:\n{}\n\n".format(
                instance.lens_galaxies, instance.source_galaxies)

    class Result(PhaseImaging.Result):
        @property
        def unmasked_lens_plane_model_image(self):
            return self.most_likely_fit.unmasked_blurred_image_plane_image_of_planes[0]

        @property
        def unmasked_source_plane_model_image(self):
            return self.most_likely_fit.unmasked_blurred_image_plane_image_of_planes[1]


class LensPlanePhase(PhaseImaging):
    """
    Fit only the lens galaxy light.
    """

    lens_galaxies = af.PhaseProperty("lens_galaxies")

    def __init__(self, phase_name, tag_phases=True, phase_folders=None,
                 lens_galaxies=None,
                 optimizer_class=af.MultiNest,
                 sub_grid_size=2, bin_up_factor=None,
                 image_psf_shape=None, mask_function=None, inner_mask_radii=None,
                 cosmology=cosmo.Planck15,
                 auto_link_priors=False):

        super(LensPlanePhase, self).__init__(phase_name=phase_name,
                                             tag_phases=tag_phases,
                                             phase_folders=phase_folders,
                                             optimizer_class=optimizer_class,
                                             sub_grid_size=sub_grid_size,
                                             bin_up_factor=bin_up_factor,
                                             image_psf_shape=image_psf_shape,
                                             mask_function=mask_function,
                                             inner_mask_radii=inner_mask_radii,
                                             cosmology=cosmology,
                                             auto_link_priors=auto_link_priors)
        self.lens_galaxies = lens_galaxies

    @property
    def uses_hyper_images(self):
        return any([galaxy.uses_hyper_images for galaxy in self.lens_galaxies])

    @property
    def uses_inversion(self):
        if self.lens_galaxies:
            for galaxy_model in self.lens_galaxies:
                if galaxy_model.pixelization is not None:
                    return True
        return False

    class Analysis(PhaseImaging.Analysis):

        def figure_of_merit_for_fit(self, tracer):
            raise NotImplementedError()

        def tracer_for_instance(self, instance):
            instance = self.associate_images(instance=instance)
            return ray_tracing.TracerImagePlane(lens_galaxies=instance.lens_galaxies,
                                                image_plane_grid_stack=self.lens_data.grid_stack,
                                                cosmology=self.cosmology)

        def padded_tracer_for_instance(self, instance):
            instance = self.associate_images(instance=instance)
            return ray_tracing.TracerImagePlane(lens_galaxies=instance.lens_galaxies,
                                                image_plane_grid_stack=self.lens_data.padded_grid_stack,
                                                cosmology=self.cosmology)

        @classmethod
        def describe(cls, instance):
            return "\nRunning lens lens for... \n\nLens Galaxy::\n{}\n\n".format(
                instance.lens_galaxies)

    class Result(PhaseImaging.Result):

        @property
        def unmasked_lens_plane_model_image(self):
            return self.most_likely_fit.unmasked_blurred_image_plane_image_of_planes[0]


class SensitivityPhase(PhaseImaging):
    lens_galaxies = af.PhaseProperty("lens_galaxies")
    source_galaxies = af.PhaseProperty("source_galaxies")
    sensitive_galaxies = af.PhaseProperty("sensitive_galaxies")

    def __init__(self, phase_name, tag_phases=None, phase_folders=None,
                 lens_galaxies=None, source_galaxies=None,
                 sensitive_galaxies=None,
                 optimizer_class=af.MultiNest, sub_grid_size=2,
                 bin_up_factor=None, mask_function=None,
                 cosmology=cosmo.Planck15):
        """
        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper
        passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        sub_grid_size: int
            The side length of the subgrid
        """

        super(SensitivityPhase, self).__init__(phase_name=phase_name,
                                               tag_phases=tag_phases,
                                               phase_folders=phase_folders,
                                               optimizer_class=optimizer_class,
                                               sub_grid_size=sub_grid_size,
                                               bin_up_factor=bin_up_factor,
                                               mask_function=mask_function,
                                               cosmology=cosmology)

        self.lens_galaxies = lens_galaxies or []
        self.source_galaxies = source_galaxies or []
        self.sensitive_galaxies = sensitive_galaxies or []

    # noinspection PyAbstractClass
    class Analysis(PhaseImaging.Analysis):

        def __init__(self, lens_data, cosmology, phase_name, results=None):
            self.lens_data = lens_data
            super(PhaseImaging.Analysis, self).__init__(cosmology=cosmology,
                                                        results=results)

        def fit(self, instance):
            """
            Determine the fit of a lens galaxy and source galaxy to the lens_data in this lens.

            Parameters
            ----------
            instance
                A model instance with attributes

            Returns
            -------
            fit: Fit
                A fractional value indicating how well this model fit and the model lens_data itself
            """
            tracer_normal = self.tracer_normal_for_instance(instance)
            tracer_sensitive = self.tracer_sensitive_for_instance(instance)
            fit = self.fit_for_tracers(tracer_normal=tracer_normal,
                                       tracer_sensitive=tracer_sensitive)
            return fit.figure_of_merit

        def visualize(self, instance, image_path, during_analysis):
            self.plot_count += 1

            tracer_normal = self.tracer_normal_for_instance(instance)
            tracer_sensitive = self.tracer_sensitive_for_instance(instance)
            fit = self.fit_for_tracers(tracer_normal=tracer_normal,
                                       tracer_sensitive=tracer_sensitive)

            ccd_plotters.plot_ccd_subplot(ccd_data=self.lens_data.ccd_data,
                                          mask=self.lens_data.mask,
                                          positions=self.lens_data.positions,
                                          output_path=image_path, output_format='png')

            ccd_plotters.plot_ccd_individual(ccd_data=self.lens_data.ccd_data,
                                             mask=self.lens_data.mask,
                                             positions=self.lens_data.positions,
                                             output_path=image_path,
                                             output_format='png')

            ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer_normal,
                                                          output_path=image_path,
                                                          output_format='png',
                                                          output_filename='tracer_normal')

            ray_tracing_plotters.plot_ray_tracing_subplot(tracer=tracer_sensitive,
                                                          output_path=image_path,
                                                          output_format='png',
                                                          output_filename='tracer_sensitive')

            sensitivity_fit_plotters.plot_fit_subplot(fit=fit, output_path=image_path,
                                                      output_format='png')

            return fit

        def tracer_normal_for_instance(self, instance):
            return ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=instance.lens_galaxies,
                source_galaxies=instance.source_galaxies,
                image_plane_grid_stack=self.lens_data.grid_stack,
                border=self.lens_data.border)

        def tracer_sensitive_for_instance(self, instance):
            return ray_tracing.TracerImageSourcePlanes(
                lens_galaxies=instance.lens_galaxies + instance.sensitive_galaxies,
                source_galaxies=instance.source_galaxies,
                image_plane_grid_stack=self.lens_data.grid_stack,
                border=self.lens_data.border)

        def fit_for_tracers(self, tracer_normal, tracer_sensitive):
            return sensitivity_fit.fit_lens_data_with_sensitivity_tracers(
                lens_data=self.lens_data,
                tracer_normal=tracer_normal,
                tracer_sensitive=tracer_sensitive)

        @classmethod
        def describe(cls, instance):
            return "\nRunning lens/source lens for... \n\nLens Galaxy:\n{}\n\nSource " \
                   "Galaxy:\n{}\n\n Sensitive " \
                   "Galaxy\n{}\n\n ".format(instance.lens_galaxies,
                                            instance.source_galaxies,
                                            instance.sensitive_galaxies)



import numpy as np

from autolens.data.array.util import grid_util

from autolens.data.array import grids

from test.unit.mock.data import mock_mask


class MockRegularGrid(grids.RegularGrid):

    def __new__(cls, mask, pixel_scale=1.0, *args, **kwargs):

        regular_grid = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(
            mask=mask, pixel_scales=(pixel_scale, pixel_scale))

        obj = regular_grid.view(cls)
        obj.mask = mask
        obj.interpolator = None

        return obj

    def __init__(self, mask, pixel_scale=1.0):
        pass


class MockSubGrid(grids.SubGrid):

    def __new__(cls, mask, pixel_scale=1.0, sub_grid_size=2, *args, **kwargs):

        sub_grid = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=mask, pixel_scales=(pixel_scale, pixel_scale), sub_grid_size=sub_grid_size)

        obj = sub_grid.view(cls)
        obj.mask = mask
        obj.sub_grid_size = sub_grid_size
        obj.sub_grid_length = int(obj.sub_grid_size ** 2.0)
        obj.sub_grid_fraction = 1.0 / obj.sub_grid_length
        obj.interpolator = None

        return obj

    def __init__(self, mask, pixel_scale=1.0, sub_grid_size=2):
        pass

class MockClusterGrid(grids.ClusterGrid):

    pass


class MockGridStack(grids.GridStack):

    def __init__(self, regular, sub, blurring, pixelization=None):

        super(MockGridStack, self).__init__(regular=regular, sub=sub, blurring=blurring, pixelization=pixelization)


class MockPaddedRegularGrid(grids.PaddedRegularGrid):

    def __new__(cls, image_shape, psf_shape, pixel_scale=1.0, *args, **kwargs):

        padded_shape = (image_shape[0] + psf_shape[0] - 1, image_shape[1] + psf_shape[1] - 1)

        padded_mask = mock_mask.MockMask(array=np.full(fill_value=False, shape=padded_shape), pixel_scale=pixel_scale)

        regular_grid = grid_util.regular_grid_1d_from_shape_pixel_scales_and_origin(
            shape=padded_mask.shape, pixel_scales=(pixel_scale, pixel_scale), origin=(0.0, 0.0))

        obj = regular_grid.view(cls)
        obj.image_shape = (5, 5)
        obj.mask = padded_mask
        obj.interpolator = None

        return obj

    def __init__(self, image_shape, psf_shape):
        pass


class MockPaddedSubGrid(grids.PaddedSubGrid):

    def __new__(cls, image_shape, psf_shape, pixel_scale=1.0, sub_grid_size=2, *args, **kwargs):

        padded_shape = (image_shape[0] + psf_shape[0] - 1, image_shape[1] + psf_shape[1] - 1)

        padded_mask = mock_mask.MockMask(array=np.full(fill_value=False, shape=padded_shape), pixel_scale=pixel_scale)

        sub_grid = grid_util.sub_grid_1d_from_shape_pixel_scales_and_sub_grid_size(
            shape=padded_mask.shape, pixel_scales=(pixel_scale, pixel_scale), sub_grid_size=sub_grid_size)

        obj = sub_grid.view(cls)
        obj.image_shape = image_shape
        obj.mask = padded_mask
        obj.sub_grid_size = 2
        obj.sub_grid_length = int(obj.sub_grid_size ** 2.0)
        obj.sub_grid_fraction = 1.0 / obj.sub_grid_length
        obj.interpolator = None

        return obj

    def __init__(self, image_shape, psf_shape):
        pass


class MockPaddedGridStack(grids.GridStack):

    def __init__(self, regular, sub):

        super(MockPaddedGridStack, self).__init__(regular=regular, sub=sub, blurring=np.array([[0.0, 0.0]]))


class MockPixSubGrid(np.ndarray):

    def __new__(cls, sub_grid, *args, **kwargs):
        return sub_grid.view(cls)

    def __init__(self, sub_grid, sub_to_regular, sub_grid_size):
        # noinspection PyArgumentList
        super().__init__()
        self.sub_grid_coords = sub_grid
        self.sub_to_regular = sub_to_regular
        self.total_pixels = sub_to_regular.shape[0]
        self.sub_grid_size = sub_grid_size
        self.sub_grid_length = int(sub_grid_size ** 2.0)
        self.sub_grid_fraction = 1.0 / self.sub_grid_length


class MockPixGridStack(object):

    def __init__(self, regular, sub, blurring=None, pix=None, regular_to_pixelization=None):
        self.regular = grids.RegularGrid(regular, mask=None)
        self.sub = sub
        self.blurring = grids.RegularGrid(blurring, mask=None) if blurring is not None else None
        self.pixelization = grids.PixelizationGrid(pix, regular_to_pixelization=regular_to_pixelization,
                                                   mask=None) if pix is not None else np.array([[0.0, 0.0]])


class MockBorders(np.ndarray):

    def __new__(cls, arr=np.array([0]), *args, **kwargs):
        """The borders of a regular grid, containing the pixel-index's of all masked pixels that are on the \
        mask's border (e.g. they are next to a *True* value in at least one of the surrounding 8 pixels and at one of \
        the exterior edge's of the mask).

        This is used to relocate demagnified pixel's in a grid to its border, so that they do not disrupt an \
        adaptive pixelization's inversion.

        Parameters
        -----------
        arr : ndarray
            A 1D array of the integer indexes of an *RegularGrid*'s borders pixels.
        """
        border = arr.view(cls)
        return border


from autolens.lens.plotters import lens_fit_plotters
from test.fixtures import *


@pytest.fixture(name='lens_fit_plotter_path')
def make_lens_fit_plotter_setup():
    return "{}/../../test_files/plotting/fit/".format(os.path.dirname(os.path.realpath(__file__)))

def test__fit_sub_plot_lens_only(lens_fit_x1_plane_5x5, lens_fit_plotter_path, plot_patch):
    lens_fit_plotters.plot_fit_subplot(fit=lens_fit_x1_plane_5x5, should_plot_mask=True, extract_array_from_mask=True,
                                       zoom_around_mask=True, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                       output_path=lens_fit_plotter_path, output_format='png')
    assert lens_fit_plotter_path + 'lens_fit.png' in plot_patch.paths


def test__fit_sub_plot_source_and_lens(lens_fit_x2_plane_5x5, lens_fit_plotter_path, plot_patch):
    lens_fit_plotters.plot_fit_subplot(fit=lens_fit_x2_plane_5x5, should_plot_mask=True, extract_array_from_mask=True,
                                       zoom_around_mask=True, cb_tick_values=[1.0], cb_tick_labels=['1.0'],
                                       output_path=lens_fit_plotter_path, output_format='png')
    assert lens_fit_plotter_path + 'lens_fit.png' in plot_patch.paths


def test__fit_individuals__lens_only__depedent_on_input(lens_fit_x1_plane_5x5, lens_fit_plotter_path, plot_patch):

    lens_fit_plotters.plot_fit_individuals(
        fit=lens_fit_x1_plane_5x5,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_model_image=True,
        should_plot_chi_squared_map=True,
        output_path=lens_fit_plotter_path, output_format='png')

    assert lens_fit_plotter_path + 'fit_image.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_noise_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_signal_to_noise_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_model_image.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_residual_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_chi_squared_map.png' in plot_patch.paths


def test__fit_individuals__source_and_lens__depedent_on_input(lens_fit_x2_plane_5x5,
                                                               lens_fit_plotter_path, plot_patch):

    lens_fit_plotters.plot_fit_individuals(
        fit=lens_fit_x2_plane_5x5,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_model_image=True,
        should_plot_lens_subtracted_image=True,
        should_plot_source_model_image=True,
        should_plot_chi_squared_map=True,
        output_path=lens_fit_plotter_path, output_format='png')

    assert lens_fit_plotter_path + 'fit_image.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_noise_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_signal_to_noise_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_model_image.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_lens_subtracted_image.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_lens_plane_model_image.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_source_plane_model_image.png' in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_residual_map.png' not in plot_patch.paths

    assert lens_fit_plotter_path + 'fit_chi_squared_map.png' in plot_patch.paths


def test__lens_fit_for_phase__source_and_lens__depedent_on_input(lens_fit_x2_plane_5x5,
                                                               lens_fit_plotter_path, plot_patch):

    lens_fit_plotters.plot_lens_fit_for_phase(
        fit=lens_fit_x2_plane_5x5, during_analysis=False, should_plot_mask=True, positions=None,
        extract_array_from_mask=True, zoom_around_mask=True, units='arcsec',
        should_plot_image_plane_pix=True,
        should_plot_as_subplot=True,
        should_plot_all_at_end_png=False,
        should_plot_all_at_end_fits=False,
        should_plot_image=True,
        should_plot_noise_map=False,
        should_plot_signal_to_noise_map=False,
        should_plot_model_image=True,
        should_plot_lens_model_image=False,
        should_plot_lens_subtracted_image=True,
        should_plot_source_model_image=True,
        should_plot_source_plane_image=False,
        should_plot_residual_map=False,
        should_plot_chi_squared_map=True,
        should_plot_regularization_weights=False,
        visualize_path=lens_fit_plotter_path)

    assert lens_fit_plotter_path + 'lens_fit.png' in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_image.png' in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_noise_map.png' not in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_signal_to_noise_map.png' not in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_model_image.png' in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_lens_subtracted_image.png' in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_lens_plane_model_image.png' not in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_source_plane_model_image.png' in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_residual_map.png' not in plot_patch.paths
    assert lens_fit_plotter_path + 'fit_chi_squared_map.png' in plot_patch.paths



from os import path

import numpy as np
import pytest

import autofit as af
from autolens.model.galaxy import galaxy as g
from autolens.model.galaxy import galaxy_data as gd
from autolens.model.galaxy import galaxy_fit
from autolens.model.profiles import light_and_mass_profiles as lmp
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from test.unit.mock.data import mock_ccd
from test.unit.mock.data import mock_convolution
from test.unit.mock.data import mock_grids
from test.unit.mock.data import mock_mask
from test.unit.mock.lens import mock_lens_data

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(autouse=True)
def set_config_path():
    af.conf.instance = af.conf.Config(path.join(directory, "test_files/config"),
                                      path.join(directory, "output"))


#
# DATA #
#
# CCD #

@pytest.fixture(name="image_5x5")
def make_image_5x5():
    return mock_ccd.MockImage(shape=(5, 5), value=1.0)


@pytest.fixture(name="psf_3x3")
def make_psf_3x3():
    return mock_ccd.MockPSF(shape=(3, 3), value=1.0)


@pytest.fixture(name="noise_map_5x5")
def make_noise_map_5x5():
    return mock_ccd.MockNoiseMap(shape=(5, 5), value=2.0)


@pytest.fixture(name="background_noise_map_5x5")
def make_background_noise_map_5x5():
    return mock_ccd.MockBackgroundNoiseMap(shape=(5, 5), value=3.0)


@pytest.fixture(name="poisson_noise_map_5x5")
def make_poisson_noise_map_5x5():
    return mock_ccd.MockPoissonNoiseMap(shape=(5, 5), value=4.0)


@pytest.fixture(name="exposure_time_map_5x5")
def make_exposure_time_map_5x5():
    return mock_ccd.MockExposureTimeMap(shape=(5, 5), value=5.0)


@pytest.fixture(name="background_sky_map_5x5")
def make_background_sky_map_5x5():
    return mock_ccd.MockBackgrondSkyMap(shape=(5, 5), value=6.0)


@pytest.fixture(name="positions_5x5")
def make_positions_5x5():
    positions = [[[0.1, 0.1], [0.2, 0.2]], [[0.3, 0.3]]]
    return list(map(lambda position_set: np.asarray(position_set), positions))


@pytest.fixture(name="ccd_data_5x5")
def make_ccd_data_5x5(image_5x5, psf_3x3, noise_map_5x5, background_noise_map_5x5,
                      poisson_noise_map_5x5,
                      exposure_time_map_5x5, background_sky_map_5x5):
    return mock_ccd.MockCCDData(
        image=image_5x5, pixel_scale=image_5x5.pixel_scale, psf=psf_3x3,
        noise_map=noise_map_5x5,
        background_noise_map=background_noise_map_5x5,
        poisson_noise_map=poisson_noise_map_5x5,
        exposure_time_map=exposure_time_map_5x5,
        background_sky_map=background_sky_map_5x5, name='mock_ccd_data_5x5')


@pytest.fixture(name="ccd_data_6x6")
def make_ccd_data_6x6():
    image = mock_ccd.MockImage(shape=(6, 6), value=1.0)
    psf = mock_ccd.MockPSF(shape=(3, 3), value=1.0)
    noise_map = mock_ccd.MockNoiseMap(shape=(6, 6), value=2.0)
    background_noise_map = mock_ccd.MockBackgroundNoiseMap(shape=(6, 6), value=3.0)
    poisson_noise_map = mock_ccd.MockPoissonNoiseMap(shape=(6, 6), value=4.0)
    exposure_time_map = mock_ccd.MockExposureTimeMap(shape=(6, 6), value=5.0)
    background_sky_map = mock_ccd.MockBackgrondSkyMap(shape=(6, 6), value=6.0)

    return mock_ccd.MockCCDData(image=image, pixel_scale=1.0, psf=psf,
                                noise_map=noise_map,
                                background_noise_map=background_noise_map,
                                poisson_noise_map=poisson_noise_map,
                                exposure_time_map=exposure_time_map,
                                background_sky_map=background_sky_map,
                                name='mock_ccd_data_6x6')


# MASK #


@pytest.fixture(name="mask_5x5")
def make_mask_5x5():
    array = np.array([[True, True, True, True, True],
                      [True, False, False, False, True],
                      [True, False, False, False, True],
                      [True, False, False, False, True],
                      [True, True, True, True, True]])

    return mock_mask.MockMask(array=array)


@pytest.fixture(name="mask_5x5_1_pix")
def make_mask_5x5_1_pix():
    array = np.array([[True, True, True, True, True],
                      [True, True, True, True, True],
                      [True, True, False, True, True],
                      [True, True, True, True, True],
                      [True, True, True, True, True]])

    return mock_mask.MockMask(array=array)


@pytest.fixture(name="blurring_mask_5x5")
def make_blurring_mask_5x5():
    array = np.array([[False, False, False, False, False],
                      [False, True, True, True, False],
                      [False, True, True, True, False],
                      [False, True, True, True, False],
                      [False, False, False, False, False]])

    return mock_mask.MockMask(array=array)


@pytest.fixture(name="padded_mask_7x7")
def make_padded_mask_7x7():
    array = np.full(fill_value=False, shape=(7, 7))

    return mock_mask.MockMask(array=array)


@pytest.fixture(name="mask_6x6")
def make_mask_6x6():
    array = np.array([[True, True, True, True, True, True],
                      [True, True, True, True, True, True],
                      [True, True, False, False, True, True],
                      [True, True, False, False, True, True],
                      [True, True, True, True, True, True],
                      [True, True, True, True, True, True]])

    return mock_mask.MockMask(array=array)


# MASKED DATA #

@pytest.fixture(name="image_1d_5x5")
def make_image_1d_5x5(image_5x5, mask_5x5):
    return mask_5x5.map_2d_array_to_masked_1d_array(array_2d=image_5x5)


@pytest.fixture(name="noise_map_1d_5x5")
def make_noise_map_1d_5x5(noise_map_5x5, mask_5x5):
    return mask_5x5.map_2d_array_to_masked_1d_array(array_2d=noise_map_5x5)


# GRIDS #


@pytest.fixture(name="regular_grid_5x5")
def make_regular_grid_5x5(mask_5x5):
    return mock_grids.MockRegularGrid(mask=mask_5x5)


@pytest.fixture(name="sub_grid_5x5")
def make_sub_grid_5x5(mask_5x5):
    return mock_grids.MockSubGrid(mask=mask_5x5)


@pytest.fixture(name="blurring_grid_5x5")
def make_blurring_grid_5x5():
    blurring_mask = np.array([[False, False, False, False, False],
                              [False, True, True, True, False],
                              [False, True, True, True, False],
                              [False, True, True, True, False],
                              [False, False, False, False, False]])

    return mock_grids.MockRegularGrid(mask=blurring_mask)

@pytest.fixture(name="cluster_grid_5x5")
def make_cluster_grid_5x5(mask_5x5):
    return mock_grids.MockClusterGrid.from_mask_and_cluster_pixel_scale(
        mask=mask_5x5, cluster_pixel_scale=mask_5x5.pixel_scale)

@pytest.fixture(name="grid_stack_5x5")
def make_grid_stack_5x5(regular_grid_5x5, sub_grid_5x5, blurring_grid_5x5):
    return mock_grids.MockGridStack(regular=regular_grid_5x5, sub=sub_grid_5x5,
                                    blurring=blurring_grid_5x5)


@pytest.fixture(name="grid_stack_simple")
def make_grid_stack_simple(regular_grid_5x5, sub_grid_5x5, blurring_grid_5x5):
    # Manually overwrite some sub-grid and blurring grid coodinates for easier deflection angle calculations

    grid_stack = mock_grids.MockGridStack(regular=regular_grid_5x5, sub=sub_grid_5x5,
                                          blurring=blurring_grid_5x5)

    grid_stack.regular[0] = np.array([1.0, 1.0])
    grid_stack.sub[0] = np.array([1.0, 1.0])
    grid_stack.sub[1] = np.array([1.0, 0.0])
    grid_stack.sub[2] = np.array([1.0, 1.0])
    grid_stack.sub[3] = np.array([1.0, 0.0])
    grid_stack.blurring[0] = np.array([1.0, 0.0])

    return grid_stack


@pytest.fixture(name="padded_regular_grid_5x5")
def make_padded_regular_grid_5x5():
    return mock_grids.MockPaddedRegularGrid(image_shape=(5, 5), psf_shape=(3, 3))


@pytest.fixture(name="padded_sub_grid_5x5")
def make_padded_sub_grid_5x5():
    return mock_grids.MockPaddedSubGrid(image_shape=(5, 5), psf_shape=(3, 3))


@pytest.fixture(name="padded_grid_stack_5x5")
def make_padded_grid_stack_5x5(padded_regular_grid_5x5, padded_sub_grid_5x5):
    return mock_grids.MockPaddedGridStack(regular=padded_regular_grid_5x5, sub=padded_sub_grid_5x5)


# BORDERS #

@pytest.fixture(name="border_5x5")
def make_border_5x5():
    return mock_grids.MockBorders(arr=np.array([0, 1, 2, 3, 5, 6, 7, 8]))


# CONVOLVERS #


@pytest.fixture(name="convolver_image_5x5")
def make_convolver_image_5x5(mask_5x5, blurring_mask_5x5, psf_3x3):
    return mock_convolution.MockConvolverImage(mask=mask_5x5, blurring_mask=blurring_mask_5x5, psf=psf_3x3)


@pytest.fixture(name="convolver_mapping_matrix_5x5")
def make_convolver_mapping_matrix_5x5(mask_5x5, psf_3x3):
    return mock_convolution.MockConvolverMappingMatrix(mask=mask_5x5, psf=psf_3x3)


#
# MODEL #
#

# PROFILES #


@pytest.fixture(name="lp_0")
def make_lp_0():
    # noinspection PyTypeChecker
    return lp.SphericalSersic(intensity=1.0, effective_radius=2.0, sersic_index=2.0)


@pytest.fixture(name="lp_1")
def make_lp_1():
    # noinspection PyTypeChecker
    return lp.SphericalSersic(intensity=2.0, effective_radius=2.0, sersic_index=2.0)


@pytest.fixture(name="mp_0")
def make_mp_0():
    # noinspection PyTypeChecker
    return mp.SphericalIsothermal(einstein_radius=1.0)


@pytest.fixture(name="mp_1")
def make_mp_1():
    # noinspection PyTypeChecker
    return mp.SphericalIsothermal(einstein_radius=2.0)


@pytest.fixture(name="lmp_0")
def make_lmp_0():
    return lmp.EllipticalSersicRadialGradient()


# GALAXY #


@pytest.fixture(name="gal_x1_lp")
def make_gal_x1_lp(lp_0):
    return g.Galaxy(redshift=0.5, light_profile_0=lp_0)


@pytest.fixture(name="gal_x2_lp")
def make_gal_x2_lp(lp_0, lp_1):
    return g.Galaxy(redshift=0.5, light_profile_0=lp_0, light_profile_1=lp_1)


@pytest.fixture(name="gal_x1_mp")
def make_gal_x1_mp(mp_0):
    return g.Galaxy(redshift=0.5, mass_profile_0=mp_0)


@pytest.fixture(name="gal_x2_mp")
def make_gal_x2_mp(mp_0, mp_1):
    return g.Galaxy(redshift=0.5, mass_profile_0=mp_0, mass_profile_1=mp_1)


@pytest.fixture(name="gal_x1_lp_x1_mp")
def make_gal_x1_lp_x1_mp(lp_0, mp_0):
    return g.Galaxy(redshift=0.5, light_profile_0=lp_0, mass_profile_0=mp_0)


# GALAXY DATA #


@pytest.fixture(name="gal_data_5x5")
def make_gal_data_5x5(image_5x5, noise_map_5x5):
    return gd.GalaxyData(image=image_5x5, noise_map=noise_map_5x5, pixel_scale=image_5x5.pixel_scale)


@pytest.fixture(name="gal_fit_data_5x5_intensities")
def make_gal_fit_data_5x5_intensities(gal_data_5x5, mask_5x5):
    return gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2, use_intensities=True)


@pytest.fixture(name="gal_fit_data_5x5_convergence")
def make_gal_fit_data_5x5_convergence(gal_data_5x5, mask_5x5):
    return gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2, use_convergence=True)


@pytest.fixture(name="gal_fit_data_5x5_potential")
def make_gal_fit_data_5x5_potential(gal_data_5x5, mask_5x5):
    return gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2, use_potential=True)


@pytest.fixture(name="gal_fit_data_5x5_deflections_y")
def make_gal_fit_data_5x5_deflections_y(gal_data_5x5, mask_5x5):
    return gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2, use_deflections_y=True)


@pytest.fixture(name="gal_fit_data_5x5_deflections_x")
def make_gal_fit_data_5x5_deflections_x(gal_data_5x5, mask_5x5):
    return gd.GalaxyFitData(galaxy_data=gal_data_5x5, mask=mask_5x5, sub_grid_size=2, use_deflections_x=True)


# GALAXY FIT #


@pytest.fixture(name="gal_fit_5x5_intensities")
def make_gal_fit_5x5_intensities(gal_fit_data_5x5_intensities, gal_x1_lp):
    return galaxy_fit.GalaxyFit(galaxy_data=gal_fit_data_5x5_intensities, model_galaxies=[gal_x1_lp])


@pytest.fixture(name="gal_fit_5x5_convergence")
def make_gal_fit_5x5_convergence(gal_fit_data_5x5_convergence, gal_x1_mp):
    return galaxy_fit.GalaxyFit(galaxy_data=gal_fit_data_5x5_convergence, model_galaxies=[gal_x1_mp])


@pytest.fixture(name="gal_fit_5x5_potential")
def make_gal_fit_5x5_potential(gal_fit_data_5x5_potential, gal_x1_mp):
    return galaxy_fit.GalaxyFit(galaxy_data=gal_fit_data_5x5_potential, model_galaxies=[gal_x1_mp])


@pytest.fixture(name="gal_fit_5x5_deflections_y")
def make_gal_fit_5x5_deflections_y(gal_fit_data_5x5_deflections_y, gal_x1_mp):
    return galaxy_fit.GalaxyFit(galaxy_data=gal_fit_data_5x5_deflections_y, model_galaxies=[gal_x1_mp])


@pytest.fixture(name="gal_fit_5x5_deflections_x")
def make_gal_fit_5x5_deflections_x(gal_fit_data_5x5_deflections_x, gal_x1_mp):
    return galaxy_fit.GalaxyFit(galaxy_data=gal_fit_data_5x5_deflections_x, model_galaxies=[gal_x1_mp])

############
### LENS ###
############

##### Lens Data ###

@pytest.fixture(name="lens_data_5x5")
def make_lens_data_5x5(ccd_data_5x5, mask_5x5, grid_stack_5x5, padded_grid_stack_5x5, border_5x5,
                       convolver_image_5x5, convolver_mapping_matrix_5x5, cluster_grid_5x5):
    return mock_lens_data.MockLensData(
        ccd_data=ccd_data_5x5, mask=mask_5x5, grid_stack=grid_stack_5x5, padded_grid_stack=padded_grid_stack_5x5,
        border=border_5x5, convolver_image=convolver_image_5x5, convolver_mapping_matrix=convolver_mapping_matrix_5x5,
        cluster=cluster_grid_5x5)


### Plane ####

from autolens.lens import plane as pl


@pytest.fixture(name='plane_5x5')
def make_plane_5x5(gal_x1_lp, grid_stack_5x5):
    return pl.Plane(galaxies=[gal_x1_lp], grid_stack=grid_stack_5x5,
                    compute_deflections=False)


### Ray Tracing ####

from autolens.lens import ray_tracing


@pytest.fixture(name='tracer_x1_plane_5x5')
def make_tracer_x1_plane_5x5(gal_x1_lp, grid_stack_5x5):
    return ray_tracing.TracerImagePlane(lens_galaxies=[gal_x1_lp], image_plane_grid_stack=grid_stack_5x5)

@pytest.fixture(name='tracer_x2_plane_5x5')
def make_tracer_x2_plane_5x5(gal_x1_lp, gal_x1_mp, grid_stack_5x5):
    return ray_tracing.TracerImageSourcePlanes(lens_galaxies=[gal_x1_mp, gal_x1_lp],
                                               source_galaxies=[gal_x1_lp],
                                               image_plane_grid_stack=grid_stack_5x5)

### Lens Fit ####

from autolens.lens import lens_fit

@pytest.fixture(name='lens_fit_x1_plane_5x5')
def make_lens_fit_x1_plane_5x5(lens_data_5x5, tracer_x1_plane_5x5):
    return lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5, tracer=tracer_x1_plane_5x5)

@pytest.fixture(name='lens_fit_x2_plane_5x5')
def make_lens_fit_x2_plane_5x5(lens_data_5x5, tracer_x2_plane_5x5):
    return lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data_5x5, tracer=tracer_x2_plane_5x5)

### Sensitive Fit ###

from autolens.lens import sensitivity_fit

@pytest.fixture(name='sensitivity_fit_5x5')
def make_sensitivity_fit_5x5(lens_data_5x5):

    lens_galaxy = g.Galaxy(redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=1.0))
    lens_subhalo = g.Galaxy(redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=0.1))
    source_galaxy = g.Galaxy(redshift=0.5, light=lp.EllipticalSersic(intensity=1.0))

    tracer_normal = ray_tracing.TracerImageSourcePlanes(
        lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy], image_plane_grid_stack=lens_data_5x5.grid_stack)

    tracer_sensitivity = ray_tracing.TracerImageSourcePlanes(
        lens_galaxies=[lens_galaxy, lens_subhalo], source_galaxies=[source_galaxy],
        image_plane_grid_stack=lens_data_5x5.grid_stack)

    return sensitivity_fit.SensitivityProfileFit(lens_data=lens_data_5x5, tracer_normal=tracer_normal,
                                                 tracer_sensitive=tracer_sensitivity)

##############
## PIPELINE ##
#############

### Phase ###

from autofit.tools import pipeline as af_pipeline
from autolens.pipeline import pipeline
from autolens.pipeline.phase import phase_imaging, phase_hyper
from test.unit.mock.pipeline import mock_pipeline

@pytest.fixture(name="mask_function_5x5_1_pix")
def make_mask_function_5x5_1_pix():

    def mask_function_5x5_1_pix(image):

        array = np.array([[True, True, True, True, True],
                          [True, True, True, True, True],
                          [True, True, False, True, True],
                          [True, True, True, True, True],
                          [True, True, True, True, True]])

        return mock_mask.MockMask(array=array)

    return mask_function_5x5_1_pix

@pytest.fixture(name="mask_function_5x5")
def make_mask_function_5x5():

    def mask_function_5x5(image):

        array = np.array([[True, True, True, True, True],
                          [True, False, False, False, True],
                          [True, False, False, False, True],
                          [True, False, False, False, True],
                          [True, True, True, True, True]])

        return mock_mask.MockMask(array=array)

    return mask_function_5x5

@pytest.fixture(name="phase_5x5")
def make_phase_5x5(mask_function_5x5):

    return phase_imaging.phase_imaging.LensSourcePlanePhase(
        optimizer_class=mock_pipeline.MockNLO, mask_function=mask_function_5x5, phase_name='test_phase')

@pytest.fixture(name="results_5x5")
def make_results():
    return mock_pipeline.MockResults(model_image=np.ones((5,5)),
                       galaxy_images=[2.0*np.ones((5,5)), 3.0*np.ones((5,5))])


@pytest.fixture(name="results_collection_5x5")
def make_results_collection(results_5x5):
    results_collection = af_pipeline.ResultsCollection()
    results_collection.add("phase", results_5x5)
    return results_collection


from .phase import GalaxyFitPhase, Phase
from autolens.pipeline.phase.phase_hyper import HyperGalaxyPhase
from .phase_imaging import PhaseImaging, MultiPlanePhase, LensSourcePlanePhase, LensPlanePhase
from .phase_positions import PhasePositions
